#!/usr/bin/env python
#
# LADFIT.PY - robust linear fitting.
#

from __future__ import print_function

__authors__ = 'David Nidever <dnidever@montana.edu>'
__version__ = '20200321'  # yyyymmdd

import numpy as np
import warnings

# Ignore these warnings, it's a bug
warnings.filterwarnings("ignore", message="numpy.dtype size changed")
warnings.filterwarnings("ignore", message="numpy.ufunc size changed")


def ladmdfunc(b, x, y, eps=1e-7):
    a = np.median(y - b*x)
    d = y - (b * x + a)
    absdev = np.sum(np.abs(d))
    nz, = np.where(y != 0.0)
    nzcount = len(nz)
    if nzcount != 0:
        d[nz] = d[nz] / np.abs(y[nz]) #Normalize
    nz, = np.where(np.abs(d) > eps)
    nzcount = len(nz)
    if nzcount != 0:        #Sign fcn, +1 for d > 0, -1 for d < 0, else 0.
        return np.sum(x[nz] * ((d[nz] > 0)*1 - (d[nz] < 0)*1)), a, absdev
    else:
        return 0.0, a, absdev

#-------------------------------------------------------------------------
def ladfit(xin, yin):
    """
    Copyright (c) 1994-2015, Exelis Visual Information Solutions, Inc. All
    rights reserved. Unauthorized reproduction is prohibited.

    LADFIT

    This function fits the paired data {X(i), Y(i)} to the linear model,
    y = A + Bx, using a "robust" least absolute deviation method. The
    result is a two-element vector containing the model parameters, A
    and B.


    Result = LADFIT(X, Y)

    Parameters
    ----------
       X:    An n-element vector of type integer, float or double.

       Y:    An n-element vector of type integer, float or double.

    EXAMPLE:
       Define two n-element vectors of paired data.
         x = [-3.20, 4.49, -1.66, 0.64, -2.43, -0.89, -0.12, 1.41, $
               2.95, 2.18,  3.72, 5.26]
         y = [-7.14, -1.30, -4.26, -1.90, -6.19, -3.98, -2.87, -1.66, $
              -0.78, -2.61,  0.31,  1.74]
       Compute the model parameters, A and B.
         result = ladfit(x, y, absdev = absdev)
       The result should be the two-element vector:
         [-3.15301, 0.930440]
       The keyword parameter should be returned as:
         absdev = 0.636851

    REFERENCE:
       Numerical Recipes, The Art of Scientific Computing (Second Edition)
       Cambridge University Press, 2nd Edition.
       ISBN 0-521-43108-5
    This is adapted from the routine MEDFIT described in:
    Fitting a Line by Minimizing Absolute Deviation, Page 703.

    MODIFICATION HISTORY:
      Written by:  GGS, RSI, September 1994
      Modified:    GGS, RSI, July 1995
                    Corrected an infinite loop condition that occured when
                    the X input parameter contained mostly negative data.
      Modified:    GGS, RSI, October 1996
                    If least-absolute-deviation convergence condition is not
                    satisfied, the algorithm switches to a chi-squared model.
                    Modified keyword checking and use of double precision.
      Modified:    GGS, RSI, November 1996
                    Fixed an error in the computation of the median with
                    even-length input data. See EVEN keyword to MEDIAN.
      Modified:    DMS, RSI, June 1997
         Simplified logic, remove SIGN and MDfunc2 functions.
      Modified:    RJF, RSI, Jan 1999
         Fixed the variance computation by adding some double
         conversions.  This prevents the function from generating
         NaNs on some specific datasets (bug 11680).
       Modified: CT, RSI, July 2002: Convert inputs to float or double.
            Change constants to double precision if necessary.
       CT, March 2004: Check for quick return if we found solution.
    """
    
    nx = len(xin)
  
    if nx != len(yin):
        raise ValuError("X and Y must be vectors of equal length.")

    x = np.float64(xin)
    y = np.float64(yin)

    sx = np.sum(x)
    sy = np.sum(y)

    #  the variance computation is sensitive to roundoff, so we do this
    #  math in DP
    sxy = np.sum(x*y)
    sxx = np.sum(x*x)
    delx = nx * sxx - sx**2

    if (delx == 0.0):         #All X's are the same
        result = [np.median(y), 0.0] #Bisect the range w/ a flat line
        absdev = np.sum(np.abs(y-np.median(y)))/nx
        return np.array(result), absdev

    aa = (sxx * sy - sx * sxy) / delx #Least squares solution y = x * aa + bb
    bb = (nx * sxy - sx * sy) / delx
    chisqr = np.sum((y - (aa + bb*x))**2)
    sigb = np.sqrt(chisqr / delx)     #Standard deviation
    
    b1 = bb
    eps = 1e-7
    f1,aa,absdev = ladmdfunc(b1, x, y, eps=eps)

    # Quick return. The initial least squares gradient is the LAD solution.
    if (f1 == 0.):
        bb = b1
        absdev = absdev / nx
        return np.array([aa, bb],float), absdev

    #delb = ((f1 >= 0) ? 3.0 : -3.0) * sigb
    delb = 3.0*sigb if (f1 >= 0) else -3.0*sigb
    
    b2 = b1 + delb
    f2,aa,absdev = ladmdfunc(b2, x, y, eps=eps)

    while (f1*f2 > 0):     #Bracket the zero of the function
        b1 = b2
        f1 = f2
        b2 = b1 + delb
        f2,aa,absdev = ladmdfunc(b2, x, y, eps=eps)


    # In case we finish early.
    bb = b2
    f = f2

    #Narrow tolerance to refine 0 of fcn.
    sigb = 0.01 * sigb

    while ((np.abs(b2-b1) > sigb) and (f != 0)): #bisection of interval b1,b2.
        bb = 0.5 * (b1 + b2)
        if (bb == b1 or bb == b2):
            break
        f,aa,absdev = ladmdfunc(bb, x, y, eps=eps)
        if (f*f1 >= 0):
            f1 = f
            b1 = bb
        else:
            f2 = f
            b2 = bb

    absdev = absdev / nx

    return np.array([aa, bb],float), absdev



#!/usr/bin/env python
#
# PLOTTING.PY - plotting functions
#

from __future__ import print_function

__authors__ = 'David Nidever <dnidever@montana.edu>'
__version__ = '20200209'  # yyyymmdd

import time
import numpy as np
import warnings
from astropy.utils.exceptions import AstropyWarning
import matplotlib.pyplot as plt
import matplotlib.colors as colors
from matplotlib.backend_bases import MouseButton
from scipy import stats
import copy
from . import utils as dln
from . import ladfit

# Ignore these warnings, it's a bug
warnings.filterwarnings("ignore", message="numpy.dtype size changed")
warnings.filterwarnings("ignore", message="numpy.ufunc size changed")

def zscaling(im,contrast=0.25,nsample=50000):
    """
    This finds the IRAF display z1,z2 scalings
    for an image
    
    Parameters:
    im         2D image
    =contrast  Contrast to use (contrast=0.25 by default)
    =nsample   The number of points of the image to use for the sample
                (nsample=5e4 by default).

    Returns:
     [z1,z1]         The minimum and maximum value to use for display scaling.

    Examples:
      zcals = zscale(im)


    From IRAF display help

    If  the  contrast  is  not  zero  the  sample  pixels  are ranked in
    brightness to form the function I(i) where i  is  the  rank  of  the
    pixel  and  I is its value.  Generally the midpoint of this function
    (the median) is very near the peak of the image histogram and  there
    is  a  well defined slope about the midpoint which is related to the
    width of the histogram.  At the ends of the I(i) function there  are
    a  few very bright and dark pixels due to objects and defects in the
    field.  To determine  the  slope  a  linear  function  is  fit  with
    iterative rejection;
    
            I(i) = intercept + slope * (i - midpoint)
    
    If  more  than half of the points are rejected then there is no well
    defined slope and the full range of the sample defines  z1  and  z2.
    Otherwise  the  endpoints  of the linear function are used (provided
    they are within the original range of the sample):
    
            z1 = I(midpoint) + (slope / contrast) * (1 - midpoint)
            z2 = I(midpoint) + (slope / contrast) * (npoints - midpoint)

    The actual IRAF program is called zsc_zlimits and is in the file
    /net/astro/iraf2.12/iraf/pkg/images/tv/display/zscale.x


    By D.Nidever  Oct 2007  (using IRAF display zscale algorithm) 
    """

    if im.ndim != 2:
        raise ValueError('The input must be 2 dimensiona')
    nx,ny = im.shape
    n = nx*ny

    nsample = np.minimum(nsample, n)

    xind = np.round(np.random.random(nsample)*(nx-1)).astype(int)
    yind = np.round(np.random.random(nsample)*(ny-1)).astype(int)    

    f = im[xind,yind]
    si = np.argsort(f)
    f2 = f[si]
    x = np.arange(nsample)
    midpoint = np.round(nsample*0.5)
    zmin = np.min(f)
    zmax = np.max(f)
    zmed = np.median(f)

    #zmin = np.min(im)
    #zmax = np.max(im)
    #zmed = np.median(im)

    # Robust fitting program
    coef,dum = ladfit.ladfit(x,f2)
    coef = coef[::-1]
    #coef = dln.poly_fit(x,f2,1,robust=True)
    
    # y = m*x + b
    # I = intercept + slope * (i-midpoint)
    # I = intercept + slope * i - slope*midpoint
    # I = slope * i + (intercept - slope*midpoint)
    # b = intercept - slope*midpoint
    # intercept = b + slope*midpoint
    
    slope = coef[0]
    intercept = coef[1] + slope*midpoint

    #    z1 = I(midpoint) + (slope / contrast) * (1 - midpoint)
    #    z2 = I(midpoint) + (slope / contrast) * (npoints - midpoint)
    #z1 = f2[midpoint] + (slope/contrast) * (1L - midpoint)
    #z2 = f2[midpoint] + (slope/contrast) * (nsample - midpoint)
    z1 = zmed + (slope/contrast) * (1 - midpoint)
    z2 = zmed + (slope/contrast) * (nsample - midpoint)

    z1 = np.maximum(z1, zmin)
    z2 = np.minimum(z2, zmax)
    
    return [z1,z2]

    

def hist2d(x,y,z=None,statistic=None,xr=None,yr=None,dx=None,dy=None,nx=200,ny=200,
           zscale=None,log=None,noerase=False,zmin=None,zmax=None,center=True,force=True,cmap=None,
           xtitle=None,ytitle=None,title=None,colorlabel=None,origin='lower',aspect='auto',
           bright=None,minhue=0.0,maxhue=0.7,minbright=0.1,maxbright=0.7,saturation=0.9,save=None):
    """
    Make a 2D histogram of points.
    
    Parameters
    ----------
    x : numpy array
       Array of x-data points to plot.
    y : numpy array
       Array of y-data points to plot.
    z : numpy array, optional
       Array of z-data points to plot if we using a statistic.
    statistic : str, optional
       The statistic to use.  Default is 'count' if z is not input otherwise 'sum'.
         The options are: 'sum', 'mean', 'median', 'std', 'mad'.
    xr : list, optional
       The range of x-values to plot.  The default is [min(x),max(x)]
    yr : list, optional
       The range of y-values to plot.  The default is [min(x),max(x)]
    dx : float, optional
       The step size in the x-dimension.  Either dx+dy or nx+ny must be specified.
    dy : float, optional
       The step size in the y-dimension.
    nx : int, optional
       Number of bins in the x-dimension. Either nx+ny or dx+dy must be specified.
          Default is 200.
    ny : int, optional
       Number of bins in the z-dimension.  Default is 200.
    zscale : bool, optional
       Automatic min and max interval of statistic based on IRAF's zscale.
    log : bool, optional
       Use logarithmic scaling of the statistic.  Default is linear.
    noerase : bool, optional
       Do not erase or clear the figure window.  Default is to clear the figure.
    zmin : float, optional
       Minimum value to plot on colorbar.  Default is the minimum of the statistic.
    zmax : float, optional
       Maximum value to plot on colorbar.  Default is the maximum of the statistic.    
    force : bool, optional
       Force the xrange and yrange to be exactly as the way they are set (default).
    center : bool, optional
       The x/y values of a bin should correspond to the center.  By default they
         correspond to the bottom-left corner.
    cmap : str, optional
       The matplotlib color map to use.  Default is 'viridis'.
    xtitle : str, optional
       The label for the x-axis.  Default is 'X'.
    ytitle : str, optional
       The label for the y-axis.  Default is 'Y'.
    title : str, optional
       The plot title.  Default is statistic+'(Z)'.
    colorlabel : str, optional
       The colorbar label.  Default is statistic+'(Z)'.
    origin : str, optional
       The origin of the plot.  Default is 'lower'.
    aspect : str, optional
       The aspect ratio of the plot.  Default is 'auto'.
    save : str, optional
       Save the figure to this file.

    Returns
    -------

    Figure is plotted to the screen.

    Example
    -------

    hist2d(x,y,z,'mean')

    """

    # Getting the current figure, creating a new one if necessary
    fig = plt.gcf()
    
    if noerase is False:
        plt.clf()   # clear the plot

    # Statistic default
    if statistic is None:
        if z is None:
            statistic = 'count'
        else:
            statistic = 'sum'
        
    # Input dx/dy
    if dx is None:
        dx0 = None
    else:
        dx0 = copy.copy(dx)
    if dy is None:
        dy0 = None
    else:
        dy0 = copy.copy(dy)
        
    # Min and Max's
    xmin = np.min(x)
    xmax = np.max(x)
    ymin = np.min(y)
    ymax = np.max(y)    

    # Temporary DX and DY if not input
    # X-axis
    if (dx is None):
        # XRANGE set
        if xr is not None:
            dx = (xr[1]-x[0])/(nx-1)
        # XRANGE not set
        else:
            dx = (xmax-xmin)/(nx-1)
    # Y-axis
    if (dy is None):
        # YRANGE set
        if yr is not None:
            dy = (yr[1]-yr[0])/(ny-1)
        # YRANGE not set
        else:
            dy = (ymax-ymin)/(ny-1)

    # Axes reversed
    if xr is not None:
        # X axis reversed
        if xr[1]<xr[0]:
            xr = np.flip(xr)
            xflip = 1
    if yr is not None:
        # Y axis reversed
        if yr[1]<yr[0]:
            yr = np.flip(yr)
            yflip = 1

    # Cutting based on the xrange/yrange if set
    #   Otherwise the image size might get too large if there are
    #   points far outside the xrange/yrange
    # Xrange set
    if xr is not None:
        mask1 = ((x>=(xr[0]-dx)) & (x<=(xr[1]+dx)))
        if np.sum(mask1)>0:
            x = x[mask1]
            y = y[mask1]
            if z is not None:
                z = z[mask1]
        else:
            raise ValueError('No points left')

        # Min and Max's
        ymin = np.min(y)
        ymax = np.max(y)
        xmin = np.min(x)
        xmax = np.max(x)

        # Setting DX/DY if not input
        if dx0 is None: dx=(xmax-xmin)/(nx-1)
        if dy0 is None: dy=(ymax-ymin)/(ny-1)
    # yrange set
    if yr is not None:
        mask2 = ((y>=(yr[0]-dy)) & (y<=(yr[1]+dy)))
        if np.sum(mask2)>0:
            x = x[mask2]
            y = y[mask2]
            if z is not None:
                z = z[mask2]
        else:
            raise ValueError('No points left')

        # Min and Max's
        ymin = np.min(y)
        ymax = np.max(y)
        xmin = np.min(x)
        xmax = np.max(x)

        # Setting DX/DY if not input
        if dx0 is None: dx=(xmax-xmin)/(nx-1)
        if dy0 is None: dy=(ymax-ymin)/(ny-1)


    # Step sizes must be positive
    dx = np.abs(dx)
    dy = np.abs(dy)


    # Setting the range
    if xr is not None:
        # Default is that the xrange/yrange are for the centers
        # Move xmin/xmax accordingly
        if center is True:
            off = 0.5*dx
        else:
            off = 0.0

        # xmin and xmax must be xrange+/-integer*dx
        if force is True:
            # If xmin < xrange[0] then move it down an integer number of dx's
            if (xmin < (xr[0]-off)):
                diff = (xr[0]-off)-xmin
                xmin = (xr[0]-off)-np.ceil(diff/dx)*dx
            else:
                xmin = xr[0]-off
            # If xmax > xrange[1] then move it up an integer number of dx's
            if (xmax > (xr[1]+off)):
                diff = xmax-(xr[1]+off)
                xmax = (xr[1]+off)+np.ceil(diff/dx)*dx
            else:
                xmax = xr[1]+off

        # Don't force the xrange
        else:
            if ((x[0]-off) < xmin) | ((xr[1]+off) > xmax):
                xmin = np.minimum((xr[0]-off), xmin)
                xmax = np.maximum((xr[1]+off), xmax)

    # Setting the range
    if yr is not None:
        # Default is that the xrange/yrange are for the centers
        # Move ymin/ymax accordingly
        if center is True:
            off = 0.5*dx
        else:
            off = 0.0

        # xmin and xmax must be xrange+/-integer*dx
        if force is True:
            # If ymin < yrange[0] then move it down an integer number of dy's
            if (ymin < (yr[0]-off)):
                diff = (yr[0]-off)-ymin
                ymin = (yr[0]-off)-np.ceil(diff/dy)*dy
            else:
                ymin = yr[0]-off
            # If ymax > yrange[1] then move it up an integer number of dy's
            if (ymax > (yr[1]+off)):
                diff = ymax-(yr[1]+off)
                ymax = (yr[1]+off)+np.ceil(diff/dy)*dy
            else:
                ymax = yr[1]+off

        # Don't force the xrange
        else:
            if ((yr[0]-off) < ymin) | ((yr[1]+off) > ymax):
                ymin = np.minimum((yr[0]-off), ymin)
                ymax = np.maximum((yr[1]+off), ymax)


    # Setting final DX/DY and NX/NY
    # X-axis
    if dx0 is None:
        dx = (xmax-xmin)/(nx-1.)
    else:
        nx = int(np.floor((xmax-xmin)/dx)+1)  # only want bins fully within the range
    # Y-axis
    if dy0 is None:
        dy = (ymax-ymin)/(ny-1)
    else:
        ny = int(np.floor((ymax-ymin)/dy)+1)  # only want bins fully within the range


    # Final xrange/yrange, if not already set
    if xr is None:
        xr = [xmin,xmax]
    if yr is None:
        yr = [ymin,ymax]
        
    # No z input
    if z is None:
        im, xedges, yedges = np.histogram2d(x,y,range=[xr,yr],bins=[nx,ny])
    # Statistic using z-values
    else:
        if statistic=='avg':
            cmap = 'jet'
            ima, xedges, yedges, binnumber = stats.binned_statistic_2d(x,y,z,statistic='mean',range=[xr,yr],bins=[nx,ny])
            ima = ima.T
            if bright is None:
                imt, xedges, yedges, binnumber = stats.binned_statistic_2d(x,y,z,statistic='count',range=[xr,yr],bins=[nx,ny])
                imt = imt.T
                if log:
                    pos = (imt > 0)
                    imt[pos] = np.log10(imt[pos])
                    imt[~pos] = -10
            else:
                imt, xedges, yedges, binnumber = stats.binned_statistic_2d(x,y,bright,statistic='sum',range=[xr,yr],bins=[nx,ny])
                imt = imt.T
                if log:
                    pos = (imt > 0)
                    imt[pos] = np.log10(imt[pos])
                    imt[~pos] = -10
                    
            # convert image to RGB, using HLS
            # hue is Average (IMA)   (0-360)
            # 0-red, 120-green, 240-blue
            # brightness is Total (im) (0-1)
            #ima2 = -ima    ; (blue-green-red)
            # color_convert, hue, bright, im*0.+saturation, r, g, b, /HLS_RGB
            gooda = np.isfinite(ima)
            if zmin is None: zmin = np.min(ima[gooda])
            if zmax is None: zmax = np.max(ima[gooda])            
            #ima = -ima

            # hue goes: red, orange, yellow, green, light blue, dark blue, purple, red
            # want to stop after dark blue, around hue=0.7
            
            # hsv is hue, saturation, value or hue, saturation, brightness
            hueim = np.zeros(ima.shape,float)
            #hueim[gooda] = dln.limit(dln.scale(ima[gooda],[-zmax,-zmin],[minhue,maxhue]),minhue,maxhue)
            hueim[gooda] = dln.limit(dln.scale(ima[gooda],[zmin,zmax],[minhue,maxhue]),minhue,maxhue)
            # flip hue to get blue to red
            hueim = 1-hueim
            goodt = np.isfinite(imt)
            tmin = np.min(imt[goodt])
            tmax = np.max(imt[goodt])            
            brightim = np.zeros(imt.shape,float)
            brightim[goodt] = dln.limit(dln.scale(imt[goodt],[tmin,tmax],[minbright,maxbright]),minbright,maxbright)
            hsvim = np.zeros((*ima.shape,3),float)
            hsvim[:,:,0] = hueim
            hsvim[:,:,1] = saturation
            hsvim[:,:,2] = brightim
            rgbim = colors.hsv_to_rgb(hsvim)
            im = rgbim

            if title is None:
                title = 'mean(Z)'
            if colorlabel is None:
                colorlabel = 'mean(Z)'
            
        # Normal statistic
        else:
            im, xedges, yedges, binnumber = stats.binned_statistic_2d(x,y,z,statistic=statistic,range=[xr,yr],bins=[nx,ny])
            im = im.T

    # Plot the image
    #fig, ax = plt.subplots()
    norm = None
    if log is True and statistic != 'avg':
        norm = colors.LogNorm(vmin=zmin,vmax=zmax)
    if zscale is True:
        zmin,zmax = zscaling(im)
    if zmin is None:
        zmin = np.min(im[np.isfinite(im)])
    if zmax is None:
        zmax = np.max(im[np.isfinite(im)])        
    extent = [xr[0], xr[1], yr[0], yr[1]]
    plt.imshow(im,cmap=cmap,norm=norm,aspect=aspect,vmin=zmin,vmax=zmax,origin=origin,extent=extent)
    
    # Axis titles
    if xtitle is None:
        xtitle = 'X'
    plt.xlabel(xtitle)
    if ytitle is None:
        ytitle = 'Y'
    plt.ylabel(ytitle)        
    if title is None:
        if statistic is 'count':
            title = statistic
        else:
            title = statistic+'(Z)'
    plt.title(title)
        
    # Add the colorbar
    if colorlabel is None:
        if statistic is 'count':
            colorlabel = statistic
        else:
            colorlabel = statistic+'(Z)'
    plt.colorbar(label=colorlabel)

    # Save the figure
    if save is not None:
        plt.savefig(save,bbox_inches='tight')
    
    return


def display(im,x=None,y=None,log=False,xr=None,yr=None,noerase=False,zscale=False,zmin=None,zmax=None,
            xtitle=None,ytitle=None,title=None,origin='lower',aspect='auto',cmap=None,figure=None,
            save=None,colorlabel=None,interpolation=None):
    """
    Display an image.  The usual python convention is used where the image is [NY,NX]

    Parameters
    ----------
    im : 2D numpy array
      2D numpy array image to display.
    x : 1D numpy array,  optional
      1D array of x-values.  Default is to use np.arange(nx).
    y : 1D numpy array, optional
      1D array of y-values.  Default is to use np.arange(ny).
    log : boolean, optional
      Use logarithmic scaling.  Default is to use linear scaling.
    xr : list, optional
      X-axis range to use.  Default is to show the full image.
    yr : list, optional
      Y-axis range to use.  Default is to show the full image.
    noerase : boolean, optional
      Do not erase the plot window.  Default is to erase.
    zscale : boolean, optional
      Set image min/max values using the zscale algorithm.
    zmin : float, optional
      Minimum image value to plot.  Default is minimum of the entire image.
    zmax : float, optional
      Maximum image value to plot.  Default is maximum of the entire image.
    xtitle : str, optional
      X-axis title  Default is no title.
    ytitle : str, optional
      Y-axis title  Default is no title.
    title : str, optional
      The plot title.  Default is no title.
    origin : str, optional
      The origin.  Default is "lower".
    aspect : str, optional
      The aspect of the plot window.  Default is "auto".
    cmap : str, optional
      The color map.  Default is "viridis".
    figure : int, optional
      The figture window number.  Default is to use the current window.
    save : str, optional
       Save the figure to this file.
    colorlabel : str, optional
       The colorbar label.  Default is statistic+'(Z)'.
    interpolation : str, optional
       The type of interpolation.

    Returns
    -------
    A plot is made in the figure window.

    Example
    -------
    
    display(im,zmin=100,zmax=5000,xr=[100,300])


    """

    # Getting the current figure, creating a new one if necessary
    if figure is None:
        fig = plt.gcf()
    else:
        fig = plt.figure(figure)
        
    if noerase is False:
        plt.clf()   # clear the plot

    ny,nx = im.shape        
        
    # No X/Y inputs
    if x is None:
        x = np.arange(nx)
    if y is None:
        y = np.arange(ny)

    # Min's and Max's
    xmin = np.min(x)
    xmax = np.max(x)
    ymin = np.min(y)
    ymax = np.max(y)

    # Trim based on xrange/yrange
    if xr is not None:
        val0,xlo = dln.closest(x,xr[0])
        val1,xhi = dln.closest(x,xr[1])        
        im = im[:,xlo:xhi+1]
        x = x[xlo:xhi+1]
        nx,ny = im.shape
        xmin = np.min(x)
        xmax = np.max(x)
    if yr is not None:
        val0,ylo = dln.closest(y,yr[0])
        val1,yhi = dln.closest(y,yr[1])        
        im = im[ylo:yhi+1,:]
        y = y[ylo:yhi+1]
        nx,ny = im.shape        
        ymin = np.min(y)
        ymax = np.max(y)
        
    # Plot the image
    #fig, ax = plt.subplots()
    norm = None
    if log is True:
        norm = colors.LogNorm(vmin=zmin,vmax=zmax)
    if zscale is True:
        zmin,zmax = zscaling(im)

    # (left, right, bottom, top)
    extent = [xmin, xmax, ymin, ymax]
    plt.imshow(im,cmap=cmap,norm=norm,aspect=aspect,vmin=zmin,vmax=zmax,origin=origin,
               extent=extent,interpolation=interpolation)
        
    # Axis titles
    if xtitle is not None:
        plt.xlabel(xtitle)
    if ytitle is not None:
        plt.ylabel(ytitle)        
    if title is not None:
        plt.title(title)
        
    # Add the colorbar
    if colorlabel is None:
        colorlabel = ''
    plt.colorbar(label=colorlabel)

    # Save the figure
    if save is not None:
        plt.savefig(save,bbox_inches='tight')
    
    return


def plot(x,y=None,c=None,marker=None,fill=True,size=None,log=False,noerase=False,
         vmin=None,vmax=None,linewidth=None,xtitle=None,ytitle=None,title=None,
         xr=None,yr=None,cmap=None,alpha=None,figure=None,xflip=False,
         yflip=False,save=None,colorlabel=None):
    """
    Create a line or scatter plot.  like plotc.pro

    Parameters
    ----------
    x : numpy array
       Array of x-values to plot.
    y : numpy array
       Array of x-values to plot.
    c : numpy array, optional
       Array of values to color-code the points by.
    marker : float/int, optional
       Marker type.
    fill : boolean, optional
       Use filled markers.  Default is fill=True.
    size : float/int, optional
       Marker size.
    log : boolean, optional
       If c input, use a logarithmic scale.
    noerase : boolean, optional
       Do not erase the current plotting figure.  Default is to erase the figure.
    vmin : float, optional
       Minimum color value to plot.
    vmax : float, optional
       Maximum color value to plot.
    linewidth : float, optional
       Linewidth to use.
    xtitle : str, optional
       X-axis title.
    ytitle : str, optional
       Y-axis title.
    title : str, optional
       Main figure title.
    xr : list, optional
       X-axis minimum and maximum range.
    yr : list, optional
       Y-axis minimum and maximum range.
    cmap : str, optional
       Color map to use.
    alpha : float, optional
       Alpha value to use (between 0 and 1).
    figure : int, optional
       Figure number.  The default is to use the current figure.
    xflip : boolean, optional
       Flip the x-axis.
    yflip : boolean, optional
       Flip the y-axis.
    save : str, optional
       Save the figure to this file.
    colorlabel : str, optional
       The colorbar label.  Default is statistic+'(Z)'.

    Returns
    -------
    Figure is plotted to the screen.

    Example
    -------

    plotting(x,y)

    """

    # xerr, yerr, symbol size

    # No Y-input
    if y is None:
        y = np.array(x).copy()
        x = np.arange(len(y))
    
    # Getting the current figure, creating a new one if necessary
    if figure is None:
        fig = plt.gcf()
    else:
        fig = plt.figure(figure)
        
    if noerase is False:
        plt.clf()   # clear the plot

    if fill==False:
        if c is not None:
            edgecolor = None
        else:
            edgecolor = 'b'
        facecolor = 'none'
    else:
        edgecolor = None
        facecolor = None
        
    # Make the plot
    norm = None
    if log is True:
        norm = colors.LogNorm(vmin=vmin,vmax=vmax)
        plt.scatter(x,y,c=c,marker=marker,s=size,cmap=cmap,norm=norm,
                    alpha=alpha,edgecolor=edgecolor,facecolor=facecolor,
                    linewidth=linewidth)        
    else:
        plt.scatter(x,y,c=c,marker=marker,s=size,cmap=cmap,norm=norm,
                    edgecolor=edgecolor,facecolor=facecolor,vmin=vmin,
                    vmax=vmax,alpha=alpha,linewidth=linewidth)

    
    # Axis titles
    if xtitle is not None:
        plt.xlabel(xtitle)
    if ytitle is not None:
        plt.ylabel(ytitle)        
    if title is not None:
        plt.title(title)

    # Axis ranges
    if xr is not None:
        plt.xlim(xr)
    if yr is not None:
        plt.ylim(yr)

    # Flip axes
    if xflip:
        if xr is None:
            xr = dln.minmax(x)
            xr = [xr[0]-0.05*dln.valrange(x),xr[1]+0.05*dln.valrange(x)]            
        plt.xlim(np.flip(xr))
    if yflip:
        if yr is None:
            yr = dln.minmax(y)
            yr = [yr[0]-0.05*dln.valrange(y),yr[1]+0.05*dln.valrange(y)]  
        plt.ylim(np.flip(yr))
                         
    # Add the colorbar
    if c is not None and len(c)==len(x):
        if colorlabel is None:
            colorlabel = ''
        plt.colorbar(label=colorlabel)

    # Save the figure
    if save is not None:
        plt.savefig(save,bbox_inches='tight')
        
    return

def oplot(*args,**kwargs):
    plot(*args,**kwargs,noerase=True)

def scatter(*args,**kwargs):
    plot(*args,**kwargs)


class Cursor():

    def __init__(self):
        self.name = 'test'
        self.coords = None
        self.binding_id = None
     
def cursor():
    """
    This returns the position of a cursor click on a matplotlib figure window.

    Parameters
    ----------
    None

    Returns
    -------
    coords : list
      List of data coordinates.

    Example
    -------

    coords = cursor()

    Click in the figure window.
    data coords: 2.981026 5.637763

    print(coords)
    [2.981026, 5.637763]

    Note, the coords variable will be a blank list until the click in the figure window.

    """

    fig = plt.gcf()

    curs.coords = []
    curs.binding_id = None

    
    def on_click(event):
        if event.button is MouseButton.LEFT:
            x, y = event.x, event.y
            if event.inaxes:
                ax = event.inaxes  # the axes instance
                print('data coords: %f %f' % (event.xdata, event.ydata))
                curs.coords += [event.xdata,event.ydata]
            plt.disconnect(curs.binding_id)
            
    binding_id = plt.connect('button_press_event', on_click)
    curs.binding_id = binding_id

    return curs.coords


def curpdiff(spherical=False,arcsec=False):
    """
    This returns the difference of two clicked positions on a figure window.

    Parameters
    ----------
    spherical : boolean, optional
       Use spherical coordinates.  Default is False.
    arcsec : boolean, optional
       User arcseconds for spherical units.  Default is degrees.

    Returns
    -------
    data : dict
      Dictionary of data values

    Example
    -------

    data = curpdiff()

    First Click

    data coords: 2.001687 1.907833
    Second Click
    data coords: 7.058273 7.033132
    Distance = 7.1998 pixels
    Delta X = 5.0566 pixels
    Delta Y = 5.1253 pixels
    Angle = 45.3867 (CCW from Right)

    Note, the data variable will be a blank dict until the clicks in the figure window.

    """

    fig = plt.gcf()

    curs.coords = {}
    curs.binding_id = None
    print('First Click')
    
    def on_click(event):
        if event.button is MouseButton.LEFT:
            if event.inaxes:
                ax = event.inaxes  # the axes instance
                if len(curs.coords)==0:
                    print('data coords: %f %f' % (event.xdata, event.ydata))
                    curs.coords['x1'] = event.xdata
                    curs.coords['y1'] = event.ydata                    
                    curs.time = time.time()
                    print('Second Click')
                elif len(curs.coords)==2:
                    if time.time()-curs.time > 0.1:
                        print('data coords: %f %f' % (event.xdata, event.ydata))
                        curs.coords['x2'] = event.xdata
                        curs.coords['y2'] = event.ydata                                            
                        plt.disconnect(curs.binding_id)
                        
            if len(curs.coords)==4:
                x1,y1 = curs.coords['x1'],curs.coords['y1']
                x2,y2 = curs.coords['x2'],curs.coords['y2']                
                # Spherical ra/dec coordinates (or similar)
                #  use cos(dec) correction 
                if spherical:
                    mndec = np.mean([y1,y2])
                    cosdec = np.cos(np.deg2rad(mndec))
                    # Convert to arc seconds
                    if arsec:
                        mfac = 3600
                        units = 'arcsec'
                    else:
                        units = 'deg'
                else:
                    cosdec = 1.0
                    mfac = 1.0
                    units = 'pixels'
                dist = np.sqrt(((x1-x2)*cosdec)**2 + (y1-y2)**2)*mfac
                deltax = (x2-x1)*mfac*cosdec
                deltay = (y2-y1)*mfac
                angle = np.rad2deg(np.arctan2(y2-y1,x2-x1))
                print('Distance = %.4f %s' % (dist,units))
                print('Delta X = %.4f %s' % (deltax,units))
                print('Delta Y = %.4f %s' % (deltay,units))
                print('Angle = %.4f %s' % (angle,'(CCW from Right)'))
                slp,yoff = linear_coefficients([x1,x2],[y1,y2])

                curs.coords['dist'] = dist
                curs.coords['units'] = units
                curs.coords['deltax'] = deltax
                curs.coords['deltay'] = deltay
                curs.coords['angle'] = angle
                curs.coords['slp'] = slp
                curs.coords['yoff'] = yoff

                
    binding_id = plt.connect('button_press_event', on_click)
    curs.binding_id = binding_id

    return curs.coords

    
def clicker(over=False,connect=False):
    """
    This returns the position of mulitple cursor click on a matplotlib figure window.
    Click outside the data axes or right-click to stop.

    Parameters
    ----------
    over : boolean, optional
       Overplot the clicked position on the figure.  Default is False.
    connect : boolean, optional
       Connect the points with lines.  Default is False.

    Returns
    -------
    coords : list
      List of data coordinates.

    Example
    -------

    coords = clicker()

    <multiple clicks in the figure window.>

    data coords: 1.042335 7.838153
    data coords: 3.220864 6.093941
    data coords: 3.800472 8.213830
    data coords: 7.198178 3.759381
    data coords: 4.579946 0.807638
    data coords: 0.962389 3.866717
    data coords: 1.522011 6.845294

    <click outside data axes or right-click to stop>

    coords
    [[1.0423346449011444, 7.83815321583179],
    [3.2208636836628513, 6.093941326530615],
    [3.8004723270031215, 8.213829622758198],
    [7.198178167273673, 3.7593807977736566],
    [4.5799460197710715, 0.8076376004947439],
    [0.9623886251300728, 3.8667169140383444],
    [1.5220107635275753, 6.845294140383428]]

    Note, the coords variable will be a blank list until the click in the figure window.

    """

    fig = plt.gcf()

    curs.coords = []
    curs.binding_id = None
    
    def on_click(event):
        if event.button is MouseButton.LEFT:
            if event.inaxes:
                ax = event.inaxes  # the axes instance
                print('data coords: %f %f' % (event.xdata, event.ydata))
                curs.coords += [[event.xdata,event.ydata]]
                if over:
                    plt.scatter([event.xdata],[event.ydata],marker='+',c='r')
                if connect and len(curs.coords)>1:
                    x1,y1 = curs.coords[-2]
                    x2,y2 = curs.coords[-1]
                    plt.plot([x1,x2],[y1,y2],c='lightsalmon')
                if connect or over:
                    fig.canvas.draw()
            else:
                plt.disconnect(curs.binding_id)

        if event.button is MouseButton.RIGHT:
            plt.disconnect(curs.binding_id)
            
    binding_id = plt.connect('button_press_event', on_click)
    curs.binding_id = binding_id

    return curs.coords


def selector(xdata,ydata,over=False,verbose=True,color='r',backcolor='white'):
    """
    Allows the user to select data points by clicking on or near then in a figure window.
    Use left-button to select and right-button to de-select points.  
    Click outside the data axes to stop.

    Parameters
    ----------
    xdata : array or list
      List or array of x-data.
    ydata : array or list
      List or array of y-data.
    over : boolean, optional
      Overplot the points selected.  Default is False.
    verbose : boolean, optional
      Verbose output to the screen.  Default is True.
    color : str, optional
      The color for overplotting.  Default is 'r'.
    backcolor : str, optional
      The background for overplotting if a point is de-selected.
         Default is 'white'.

    Returns
    -------
    index : list
      Index of selected points.

    Example
    -------

    index = selector(x,y,over=True)                                                                                                        
    ------------------------------------------------
        NUM       X           Y        IND        
    ------------------------------------------------

           1       7.000       7.000       7
           2       6.000       6.000       6
           3       5.000       5.000       5
           4       8.000       8.000       8
           5       2.000       2.000       2
           6       5.000       5.000       5         DE-SELECTED
           7       6.000       6.000       6         DE-SELECTED
           8       7.000       7.000       7         DE-SELECTED
           9       1.000       1.000       1
    ------------------------------------------------

    index                                                                                                                                   
    [8, 2, 1]

    Note, the index variable will be a blank list until the clicks in the figure window.

    """

    fig = plt.gcf()

    curs.coords = []
    curs.binding_id = None
    curs.xdata = xdata
    curs.ydata = ydata
    curs.index = []
    curs.time = time.time()
    curs.count = 0

    if verbose:
        print('------------------------------------------------')
        print('      NUM       X           Y        IND        ')
        print('------------------------------------------------')

    
    def on_click(event):
        if event.button is MouseButton.LEFT:
            # Add point
            if event.inaxes:
                ax = event.inaxes  # the axes instance
                xcoord,ycoord = event.xdata,event.ydata
                if time.time()-curs.time > 0.1:
                    curs.time = time.time()
                    curs.coords += [[xcoord,ycoord]]
                    # Get the closest point
                    idx = ((curs.xdata-xcoord)**2+(curs.ydata-ycoord)**2).argmin()
                    # New point
                    if idx not in curs.index:
                        curs.count += 1                        
                        curs.index += [idx]
                        if verbose:
                            print('%8d%12.3f%12.3f%8d' % (curs.count,curs.xdata[idx],curs.ydata[idx],idx))
                        if over:
                            plt.scatter([curs.xdata[idx]],[curs.ydata[idx]],marker='D',facecolor='none',edgecolor=color)
                            fig.canvas.draw()
            else:
                plt.disconnect(curs.binding_id)
                if verbose:
                    print('------------------------------------------------')
                
        if event.button is MouseButton.RIGHT:
            # Remove point
            if event.inaxes:
                ax = event.inaxes  # the axes instance
                xcoord,ycoord = event.xdata,event.ydata
                if time.time()-curs.time > 0.1 and len(curs.index)>0:
                    curs.time = time.time()                    
                    curs.count += 1
                    # Get the closest point that was previously selected
                    idx = ((curs.xdata[curs.index]-xcoord)**2+(curs.ydata[curs.index]-ycoord)**2).argmin()
                    index,xcoo,ycoo = curs.index[idx],curs.xdata[curs.index][idx],curs.ydata[curs.index][idx]
                    # Point to remove
                    curs.index.pop(idx)
                    if verbose:
                        print('%8d%12.3f%12.3f%8d%20s' % (curs.count,xcoo,ycoo,index,'DE-SELECTED'))
                    if over:
                        plt.scatter([xcoo],[ycoo],marker='D',facecolor='none',edgecolor=backcolor)
                        fig.canvas.draw()                   

                
    binding_id = plt.connect('button_press_event', on_click)
    curs.binding_id = binding_id

    return curs.index


curs = Cursor()




#!/usr/bin/env python
#
# GALAXY_MODEL.PY - A simple galaxy model
#

from __future__ import print_function

__authors__ = 'David Nidever <dnidever@montana.edu>'
__version__ = '20200305'  # yyyymmdd

import numpy as np

# Ignore these warnings, it's a bug
warnings.filterwarnings("ignore", message="numpy.dtype size changed")
warnings.filterwarnings("ignore", message="numpy.ufunc size changed")

def xyz2lbd(x,y,z,R0=8.5):
    """ Convert galactocentric X/Y/Z coordinates to l,b,dist."""
    ll = np.atleast_1d(x).copy()*0.0
    bb = np.atleast_1d(x).copy()*0.0
    dd = np.atleast_1d(x).copy()*0.0    
    for i in range(len(ll)):
        xx = np.float64(x[i])
        yy = np.float64(y[i])
        zz = np.float64(z[i])
        rho = np.sqrt( (xx+R0)**2 + yy**2)  # distance from sun in X/Y plane

        lrad = np.arctan2(yy,xx+R0)
        
        brad = 0.5*np.pi - np.arctan2(rho,zz)      # this is more straighforward
        if brad > 0.5*np.pi:
            brad = brad-np.pi
        if brad < -0.5*np.pi:
            brad = brad+np.pi

        # This doesn't work if z=0
        #if cos(0.5*!dpi-brad) ne 0.0 then d = zz/cos(0.5*!dpi-brad)
        #if cos(0.5*!dpi-brad) eq 0.0 then d = abs(zz)
        d = np.sqrt( (xx+R0)**2 + yy**2 + zz**2 )

        b = np.rad2deg(brad)
        l = np.rad2deg(lrad)
        
        if l < 0.:
            l = l+360.
        
        ll[i] = l
        bb[i] = b
        dd[i] = d

    return ll,bb,dd

def vgsr2vhelio(vgsr,lon,lat,vcirc=240.0):
    """
    Given an input array called vhel with values of heliocentric
    radial velocities and input arrays of the same length with
    the Galactic coordinates of the object (gl,gb),
    this code calculates Vlsr and Vgsr.

    INPUTS:
    vgsr     Galactocentric Standard of Rest Radial Velocity
    lon      Galactic longitude (in degrees)
    lat      Galactic latitude (in degrees)
    =vcirc   MW circular velocity.  220 km/s by default.
   
    OUTPUTS:
    Vhelio  The heliocentric velocity in km/s.

    USAGE:
    IDL>vhelio = vgsr2vhelio(vgsr,lon,lat)

    By D.Nidever 2005
    """

    # code assumes gl & gb given in degrees.

    gl = np.deg2rad(lon)
    gb = np.deg2rad(lat)

    cgl=np.cos(gl)
    sgl=np.sin(gl)
    cgb=np.cos(gb)
    sgb=np.sin(gb)

    #  This equation takes the solar motion w.r.t. the LSR as
    #  (9,11,6) km/sec (Mihalas & Binney)
    #vlsr=vhel+((9.0d0*cgb*cgl)+(11.0d0*cgb*sgl)+(6.0d0*sgb))

    #  This equation takes the rotation velocity of the LSR to
    #  be 220 km/sec

    vlsr = vgsr-(vcirc*cgb*sgl)

    #   Using updated values from Dehnen & Binney 1998, MNRAS, 298, 387
    #   U = 10.00 +/- 0.36 km/s
    #   V = 5.25 +/- 0.62 km/s
    #   W = 7.17 +/- 0.38 km/s
    #   This is in a right-handed system
    #vlsr=vhel+((9.0d0*cgb*cgl)+(11.0d0*cgb*sgl)+(6.0d0*sgb))
    #vhel=vlsr-( (9.0d0*cgb*cgl)+(11.0d0*cgb*sgl)+(6.0d0*sgb) )
    vhel = vlsr-( (10.0*cgb*cgl)+(5.25*cgb*sgl)+(7.17*sgb) )
    
    return vhel


def vgsr2vlsr(v,l,b,dir,vcirc=240.0):
    """

    PURPOSE:
    Given an input array called vhel with values of heliocentric
    radial velocities and input arrays of the same length with
    the Galactic coordinates of the object (gl,gb),
  this code calculates Vlsr and Vgsr.

    INPUTS:
    vel     Vgsr/Vlsr velocity to convert
    l       Galacitc longitude in degrees
    b       Galactic latitutde in degrees
    dir     Direction of conversion
             dir=1    Vgsr->Vlsr
             dir=-1   Vlsr->Vgsr
    =vcirc  The rotation speed of the MW disk at the sun's
                 radius.  Vcirc=240 km/s by default.

    OUTPUTS:
    Vlsr/Vgsr velocity array

    USAGE:
    IDL>vlsr = vgsr2vlsr(vgsr,lon,lat)

    By D.Nidever Feb 2006
    """

    gl = l
    gb = b
    
    gl = np.deg2rad(gl)
    gb = np.deg2rad(gb)

    cgl = np.cos(gl)
    sgl = np.sin(gl)
    cgb = np.cos(gb)
    sgb = np.sin(gb)

    #  This equation takes the solar motion w.r.t. the LSR as
    #  (9,11,6) km/sec (Mihalas & Binney)
    #
    #vlsr=vhel+((9.0d0*cgb*cgl)+(11.0d0*cgb*sgl)+(6.0d0*sgb))


    #  This equation takes the rotation velocity of the LSR to
    #  be 220 km/sec
    #
    #vgsr=vlsr+(220.0d0*cgb*sgl)

    #vgsr=vlsr+(220.0d0*cgb*sgl)
    #vlsr=vgsr-(220.0d0*cgb*sgl)

    # input vgsr, output vlsr
    if dir == 1:
        v2 = v-(vcirc*cgb*sgl)

    # input vlsr, output vgsr
    if dir == -1:
        v2 = v+(vcirc*cgb*sgl)

    return v2


def galaxy_model(flat=False,vcirc=240.0,R0=8.5,rscale=12.5,zscale=5.0,rmin=1.5,vdisk=240.0,vdisp=0.0,rhelcut=None,nstars=100000):
    """
    Model of the proper motions and rvs of the galaxy.
    Mostly copied from ggss/ggss_model.pro

    INPUT:
    =vdisk    The circular velocity of the disk (default vdisk=vcirc)
    =vcirc    The circular velocity of the sun (default vcirc=220 km/s)
    =R0       The distance from the sun to the GC (default R0=8.5 kpc)
    =rscale   The upper radial cutoff of the disk (default rscale=12.5 kpc)
    =zscale   The z-height cutoff of the disk (default zscale=5.0 kpc)
    =vdisp    Velocity dispersion (default is vdisp=0 km/s)
    =rmin     The lower radial cutoff of the disk(default rmin=1.5 kpc)
    =rhelcut  Only keep stars within this distance of the sun.
    =nstars   The number of stars in the simulation (default nstars=100,000)
    /flat     Simulate a "flat" galaxy, i.e. no thickness, all stars at b=0.

    OUTPUT:
    x       Galactocentric X direction (in kpc)
    y       Galactocentric Y direction (in kpc)
    z       Galactocentric Z direction (in kpc)
    vx      Velocity in the galactocentric X direction (in km/s)
    vy      Velocity in the galactocentric Y direction (in km/s)
    vz      Velocity in the galactocentric Z direction (in km/s)
    l       Galactic longitude
    b       Galactic latitude
    vhel    Heliocentic radial velocity (in km/s)
    vlsr    Radial velocity wrt the local standard of rest (LSR) (in km/s)
    vgsr    Radial velocity wrt the galactic standard of rest (GSR) (in km/s)
    mul     True proper motion in galactic longitude (in mas/yr)
    mub     True proper motion in galactic latitude (in mas/yr)

    Written by D.Nidever  March 2006
    """

    #The Coordinate System (right-handed):
    # x = toward galactic center from sun
    # y = toward l=90
    #  z = toward NGP
    #  phi = 0 at x, positive toward +y (against galactic rotation)
    #  r = radial distance from GC


    # Getting the random positions
    np.random.seed(0)
    x = (np.random.rand(2*nstars)-0.5)*2.0*rscale
    y = (np.random.rand(2*nstars)-0.5)*2.0*rscale    
    # Flat galaxy
    if flat is True:
        z = np.zeros(nstars,float)
    # Galaxy extended in the Z direction
    else:
        z = np.random.randn(2*nstars)*zscale

    # Only want stars within a radius RSCALE and greater than RMIN*RSCALE 
    gd, = np.where( (np.sqrt(x**2.+y**2.) < rscale) & (np.sqrt(x**2.+y**2.) > rmin))
    x = x[gd]
    y = y[gd]
    z = z[gd]

    # RHEL CUT
    if rhelcut is not None:
        xhel = x+R0
        rhel = np.sqrt(xhel**2.+y**2.+z**2.)
        gd, = np.where(rhel <= rhelcut)
        if len(gd)==0:
            raise ValueError('No stars within '+str(rhelcut)+' kpc of the sun')
        x = x[gd]
        y = y[gd]
        z = z[gd]

    # Trim to number of stars that we want
    if len(x)>nstars:
        x = x[0:nstars]
        y = y[0:nstars]
        z = z[0:nstars]
        
    # Cylindrical and spherical coordinates
    rho = np.sqrt(x**2.+y**2.)    # distance from z axis
    gphi = np.arctan2(y,x)         # the galactic azimuthal coordinate (0-2*pi), in radians
    gtheta = np.arctan2(rho,z)     # the galactic polar coordinate (0-pi), in radians
    r = np.sqrt(x**2.+y**2.+z**2.) # radial distance, in kpc

    # Get Galactic coordinates
    l,b,d = xyz2lbd(x,y,z,R0=R0)
    
    # Use a radius-dependent rotation curve
    vcoef = np.array([0.0, 55.0, -4.0, 0.09])
    vdisk = np.poly1d(np.flip(vcoef))(rho)

    # Cartesian velocities
    #  Moving in the -gphi direction
    #  phihat = -sin(phi)*xhat + cos(phi)*yhat
    vx = (-vdisk)*(-np.sin(gphi))
    vy = (-vdisk)*np.cos(gphi)
    vz = x.copy()*0.
    # add velocity dispersion
    if vdisp > 0.0:
        np.random.seed(1)
        vx += np.random.randn(len(vx))*vdisp/np.sqrt(3.)
        vy += np.random.randn(len(vx))*vdisp/np.sqrt(3.)
        vz += np.random.randn(len(vx))*vdisp/np.sqrt(3.)
    vtot = np.sqrt(vx**2+vy**2+vz**2)     # total velocity, should be vcirc

    # *** NOW WE "OBSERVE" THE STARS ***

    # RADIAL VELOCITY
    
    #Calculating the heliocentric radial velocity
    # Vrad = Vrel dotted with unit vector from sun to the star
    # Vrad = V*cos(angle)

    ## Relative velocity of the star wrt the sun
    # THIS ASSUMES THAT THE SUN IS IN COMPLETE CIRCULAR MOTION
    vrelvec = np.zeros((len(x),3),np.float64)
    vrelvec[:,0] = vx
    vrelvec[:,1] = vy-vcirc    # the sun is moving in +y direction
    vrelvec[:,2] = vz

    # unit vector for our line-of-sight toward the star, rhat
    #  in helio cartesian coordinate, xhel=x+R0
    # rhat = (x/r)*xhat + (y/r)*yhat + (z/r)*zhat
    xhel = x+R0
    rhel = np.sqrt(xhel**2+y**2+z**2)
    rhat = np.zeros((len(x),3),np.float64)
    rhat[:,0] = xhel/rhel 
    rhat[:,1] = y/rhel
    rhat[:,2] = z/rhel

    # ****THIS IS VLSR NOT VHELIO*****

    # Take the dot product of the two vectors
    # vvec dot rhat = vrelvec[0]*rhat[0] + vrelvec[1]*rhat[1] + vrelvec[2]*rhat[2]
    vlsr = vrelvec*rhat
    vlsr = np.sum(vlsr,axis=1)

    # Getting Vgsr and Vhelio velocity
    vgsr = vgsr2vlsr(vlsr,l,b,-1)
    vhel = vgsr2vhelio(vgsr,l,b,vcirc=vcirc)


    # PROPER MOTIONS

    # Calculating the heliocentric transverse velocity and proper motion

    # heliocentric spherical coordinates
    htheta = 90-b   # 0 at NP, 180 at SP
    hphi = l

    # In the +LON direction

    # unit vector in the +LON direction at the star's position, lhat
    # lhat = -sin(hphi)*xhat + cos(hphi)*yhat
    lhat = np.zeros((len(x),3),np.float64)
    lhat[:,0] = -np.sin(np.deg2rad(hphi))
    lhat[:,1] = np.cos(np.deg2rad(hphi))
    lhat[:,2] = 0.0

    # transverse velocity in +LON direction
    #  relative velocity dot lhat
    vtanl = vrelvec*lhat
    vtanl = np.sum(vtanl,axis=1)

    # proper motion in +LON direction
    #   mu(arcsec/year) = Vtan(km/s)/(4.74*d(pc))
    mul = vtanl/(4.74*rhel*1000.0)          # in arcsec/year
    mul = mul * 1000.0                      # in mas/year

    # in +LAT direction

    # unit vector in the +LAT direction at the star's position, bhat
    # bhat = cos(htheta)*cos(hphi)*xhat + cos(htheta)*sin(hphi)*yhat - sin(htheta)*zhat
    bhat = np.zeros((len(x),3),np.float64)
    bhat[:,0] = np.cos(np.deg2rad(htheta))*np.cos(np.deg2rad(hphi))
    bhat[:,1] = np.cos(np.deg2rad(htheta))*np.sin(np.deg2rad(hphi))
    bhat[:,2] = -np.sin(np.deg2rad(htheta))

    # transverse velocity in +LAT direction
    #  relative velocity dot lhat
    vtanb = vrelvec*bhat
    vtanb = np.sum(vtanb,axis=1)

    # proper motion in +LAT direction
    #   mu(arcsec/year) = Vtan(km/s)/(4.74*d(pc))
    mub = vtanb/(4.74*rhel*1000.0)          # in arcsec/year
    mub = mub * 1000.0                      # in mas/year

    # The position angle of the proper motion, E of N
    pa = np.rad2deg(np.arctan2(mul,mub))
    bd, = np.where(pa < 0.0)
    if len(bd)>0:
        pa[bd] = pa[bd]+360.0

    # Make the output structure
    dt = np.dtype([('x',float),('y',float),('z',float),('l',float),('b',float),('d',float),('rho',float),('phi',float),
                   ('vx',float),('vy',float),('vz',float),('vtot',float),('vhelio',float),('vlsr',float),('vgsr',float),('pml',float),('pmb',float)])
    out = np.zeros(len(l),dtype=dt)
    out['x'] = x
    out['y'] = y
    out['z'] = z
    out['l'] = l
    out['b'] = b
    out['d'] = d
    out['rho'] = rho
    phi = 180-np.rad2deg(gphi)
    bd, = np.where(phi > 180)
    if len(bd)>0:
        phi[bd] -= 360       # left-handed system
    out['phi'] = phi
    out['vx'] = vx
    out['vy'] = vy
    out['vz'] = vz
    out['vtot'] = np.sqrt(vx**2+vy**2+vz**2)
    out['vhelio'] = vhel
    out['vlsr'] = vlsr
    out['vgsr'] = vgsr
    out['pml'] = mul
    out['pmb'] = mub

    return out



#!/usr/bin/env python
#
# COORDS.PY - coordinate utility functions.
#

from __future__ import print_function

__authors__ = 'David Nidever <dnidever@montana.edu>'
__version__ = '20190723'  # yyyymmdd

import numpy as np
import copy
from scipy.spatial import cKDTree
from scipy.optimize import minimize
from . import utils,ladfit
from astropy.coordinates import frame_transform_graph,SkyCoord
from astropy.coordinates.matrix_utilities import rotation_matrix, matrix_product, matrix_transpose
import astropy.coordinates as coord
import astropy.units as u

def rotsph(lon,lat,clon,clat,anode=None,reverse=False,original=False):
    '''
    This rotates a spherical coordinate system to a new pole

    I got the equations for this from the paper
    Calabretta et al. 2002, A&A, 395, 1077
    Equation 5.

    Also, see rotate_lb.pro that uses a matrix method
    and for which you can specify the equator you'd like.
    rotsph.pro is faster than rotate_lb.pro
    By default, the origin is the point where the two equators
    cross (unless =ANODE is set).
    This should give you the same result (within ~1E-10")
    rotate_lb,lon,lat,[clon,clat],[clon+90,0.],nlon,nlat

    Parameters
    ----------
    lon       Array of longitudes to be rotated
    lat       Array of latitudes to be rotated
    clon      Longitude of the new NORTH POLE in the old coordinate system
    clat      Latitude of the new NORTH POLE in the old coordinate system
    =anode    The "Ascending Node" which is the longitude (in the new
             system) of the first point where the old equator cross
             the new one.  This sets the zero-point of the new
             longitude.  By default the zero-point of the new
             coordinate system is the point where the two equators
             cross.
    /original Set the new longitude zero-point to be clon (if clat>0)
             and clon+180 (if clat<0).  This is the way it was
             originally done.  DON'T USE WITH "ANODE"
    /stp      Stop at the end of the program
    /reverse  The reverse operation.  In that case (nlon,nlat) should be input
           as (lon,lat). E.g.

           rotsph,ra,dec,cra,cdec,nlon,nlat
           rotsph,nlon,nlat,cra,cdec,nra,ndec,/reverse
           
           (ra,dec) and (nra,ndec) should be identical to 1E-10.

    Returns
    -------
    nlon  Array of rotated longitudes
    nlat  Array of rotated latitudes

    '''

    radeg = 180.0/np.pi

    alphap = np.array(clon/radeg)
    deltap = np.array(clat/radeg)
    phip = np.array(90.0/radeg)
    if original: phip = np.array(180.0/radeg)   # original way
    thetap = np.array(90.0/radeg)

    # By default the origin of the new coordinate system is the point
    # where the two equators cross for the first time
    #  Unless /original is set.  Then the zero-point is at clon
    #   (if clat>0) and clon+180 (if clat<0)

    # NORMAL
    if reverse is False:
        alpha = np.array(lon/radeg)
        delta = np.array(lat/radeg)

        # arg(x,y) but atan(y,x)
        phi = phip + np.arctan2( -np.cos(delta)*np.sin(alpha-alphap), np.sin(delta)*np.cos(deltap)- \
                                 np.cos(delta)*np.sin(deltap)*np.cos(alpha-alphap) )

        theta = np.arcsin( utils.limit((np.sin(delta)*np.sin(deltap)+np.cos(delta)*np.cos(deltap)*np.cos(alpha-alphap)),-1,1) )

        # Preparing the output
        nlon = phi*radeg
        nlat = theta*radeg

        # Ascending Node
        #  By default the origin of nlon is the point where the two equators
        #  cross the first time
        if anode is not None: nlon += anode

    # REVERSE
    else:
        phi = np.array(lon/radeg)
        theta = np.array(lat/radeg)

        # Ascending Node
        if anode is not None: phi = (lon-anode)/radeg

        # arg(x,y) but atan(y,x)
        alpha = alphap + np.arctan2( -np.cos(theta)*np.sin(phi-phip), np.sin(theta)*np.cos(deltap) - \
                                     np.cos(theta)*np.sin(deltap)*np.cos(phi-phip))
        delta = np.arcsin( np.sin(theta)*np.sin(deltap) + np.cos(theta)*np.cos(deltap)*np.cos(phi-phip) )

        # Preparing the output
        nlon = alpha*radeg
        nlat = delta*radeg

    # Want everything less than 360.0
    nlon = nlon % 360.0

    # Make negative points positive
    bd = (nlon < 0.0)
    if np.sum(bd)>0:
        nlon[bd] = nlon[bd]+360.0

    return nlon, nlat


def rotsphcen(lon,lat,clon,clat,polar=False,gnomic=False,reverse=False):
    '''
    This is very similar to rotsph.pro except that the coordinates
    input are not for the north pole but for the new equator.
    Everything is in DEGREES.
    
    Parameters
    ----------
    lon       Array of longitudes to be rotated
    lat       Array of latitudes to be rotated
    clon      Longitude of the new EQUATOR in the old coordinate system
    clat      Latitude of the new EQUATOR in the old coordinate system
    /polar    Return polar coordinates (rad,phi) instead of LON/LAT.
    phi starts at North.
    /gnomic   Also do a gnomic (tangent plane) projection.
    /reverse  The reverse operation.  In that case (nlon,nlat) should be input
    as (lon,lat). E.g.

           rotsphcen,ra,dec,cra,cdec,nlon,nlat
           rotsphcen,nlon,nlat,cra,cdec,nra,ndec,/reverse
           
           (ra,dec) and (nra,ndec) should be identical to 1E-10.

    Returns
    -------
    nlon  Array of rotated longitudes.  If /polar then this is PHI
       the polar angle (measured from N toward E).
       
    nlat  Array of rotated latitudes.  If /polar then this is RAD
       the polar radial distance.
    '''

    radeg = 180.0/np.pi

    # NOT polar coordinates
    if (polar is False) and (gnomic is False):

        # Get coordinates for the north pole
        np_lon = np.array(clon)
        np_lat = np.array(clat+90.0)
        if (np_lat > 90.0):
            np_lon = np.array(clon+180.0)
            np_lon = np_lon % 360.0
            np_lat = 90.0-clat

        # Run rotsph.pro
        # NORMAL
        if reverse is False:
            nlon, nlat = rotsph(lon,lat,np_lon,np_lat,original=True)
        # REVERSE
        else:
            nlon, nlat = rotsph(lon,lat,np_lon,np_lat,reverse=True,original=True)

            # need to flip them around by 180 deg b/c the zero-point
            #  is set by the NP lon
            if ((clat+90.0) > 90.0):
                nlon = (nlon+180.0) % 360.0
                nlat = -nlat

        # Make the longitudes continuous
        nlon = (nlon+180.0) % 360.0
        nlon = nlon-180.0


    # POLAR or GNOMIC
    else:

        # Making polar coordinates

        #-----------------
        # NORMAL
        #------------------
        if reverse is False:
            # Run rotsph.pro and specify the center of the field (the origin) as the
            #  the Npole
            phi, theta = rotsph(lon,lat,clon,clat,original=True)
            # phi is now going clockwise and from South
            orig_phi = phi
            phi = -phi+180.0      # counterclockwise
            phi = phi % 360.0
            rad = 90.0-theta

            # Making gnomic projection
            if gnomic:
                # Scale the radius
                rad = radeg * np.cos(theta/radeg)/np.sin(theta/radeg)
                # Now convert from gnomic polar to X/Y
                # phi is from N toward E
                # x = R*sin(phi)
                # y = R*cos(phi)
                nlon = rad*np.sin(phi/radeg)
                nlat = rad*np.cos(phi/radeg)

            # Output polar coordinates
            if polar:
                nlon = phi
                nlat = rad

        #-----------------
        # REVERSE
        #-----------------
        else:

            # Polar
            if polar:
                phi = lon
                rad = lat
                theta = 90.0-rad

            # Removing gnomic projection
            if gnomic:
                # Now convert from X/Y to gnomic polar
                # phi is from N toward E
                # x = R*sin(phi)
                # y = R*cos(phi)
                #nlon = rad*sin(phi/radeg)
                #nlat = rad*cos(phi/radeg)
                rad = np.sqrt(lon**2.0+lat**2.0)
                phi = radeg*np.arctan2(lon,lat)      # in degrees
                # Scale the radius
                #rad = radeg * cos(theta/radeg)/sin(theta/radeg)
                theta = radeg*np.arctan(radeg/rad)   # in degrees

            # phi is now going clockwise and from South
            phi = -phi+180.0       # reverse phi

            #Run rotsph.pro and specify the center of the field (the origin) as the
            #  the Npole
            nlon, nlat = rotsph(phi,theta,clon,clat,reverse=True,original=True)

    return nlon, nlat


def doPolygonsOverlap(xPolygon1, yPolygon1, xPolygon2, yPolygon2):
    """Returns True if two polygons are overlapping."""

    # How to determine if two polygons overlap.
    # If a vertex of one of the polygons is inside the other polygon
    # then they overlap.
    
    n1 = len(xPolygon1)
    n2 = len(xPolygon2)
    isin = False

    # If ranges don't overlap, then polygons don't overlap
    if rangeoverlap(xPolygon1,xPolygon2)==False or rangeoverlap(yPolygon1,yPolygon2)==False:
        return False
    
    # Loop through all vertices of second polygon
    for i in range(n2):
        # perform iterative boolean OR
        # if any point is inside the polygon then they overlap   
        isin = isin or isPointInPolygon(xPolygon1, yPolygon1, xPolygon2[i], yPolygon2[i])

    # Need to do the reverse as well, not the same
    for i in range(n1):
        isin = isin or isPointInPolygon(xPolygon2, yPolygon2, xPolygon1[i], yPolygon1[i])

    # Two polygons can overlap even if there are no vertices inside each other.
    # Need to check if the line segments overlap
    if isin==False:
        intersect = False
        # Add first vertex to the end
        xp1 = np.append( xPolygon1, xPolygon1[0] )
        yp1 = np.append( yPolygon1, yPolygon1[0] )
        xp2 = np.append( xPolygon2, xPolygon2[0] )
        yp2 = np.append( yPolygon2, yPolygon2[0] )
        for i in range(4):
            for j in range(4):
                intersect = intersect or doLineSegmentsIntersect(xp1[i:i+2],yp1[i:i+2],xp2[j:j+2],yp2[j:j+2])
                if intersect==True:
                    return True
        isin = isin or intersect
        
    return isin

def isPointInPolygon(xPolygon, yPolygon, xPt, yPt):
    """Returns boolean if a point is inside a polygon of vertices."""
    
    # How to tell if a point is inside a polygon:
    # Determine the change in angle made by the point and the vertices
    # of the polygon.  Add up the delta(angle)'s from the first (include
    # the first point again at the end).  If the point is inside the
    # polygon, then the total angle will be +/-360 deg.  If the point is
    # outside, then the total angle will be 0 deg.  Points on the edge will
    # outside.
    # This is called the Winding Algorithm
    # http://geomalgorithms.com/a03-_inclusion.html

    n = len(xPolygon)
    # Array for the angles
    angle = np.zeros(n)

    # add first vertex to the end
    xPolygon1 = np.append( xPolygon, xPolygon[0] )
    yPolygon1 = np.append( yPolygon, yPolygon[0] )

    wn = 0   # winding number counter

    # Loop through the edges of the polygon
    for i in range(n):
        # if edge crosses upward (includes its starting endpoint, and excludes its final endpoint)
        if yPolygon1[i] <= yPt and yPolygon1[i+1] > yPt:
            # if (P is  strictly left of E[i])    // Rule #4
            if isLeft(xPolygon1[i], yPolygon1[i], xPolygon1[i+1], yPolygon1[i+1], xPt, yPt) > 0: 
                 wn += 1   # a valid up intersect right of P.x

        # if edge crosses downward (excludes its starting endpoint, and includes its final endpoint)
        if yPolygon1[i] > yPt and yPolygon1[i+1] <= yPt:
            # if (P is  strictly right of E[i])    // Rule #4
            if isLeft(xPolygon1[i], yPolygon1[i], xPolygon1[i+1], yPolygon1[i+1], xPt, yPt) < 0: 
                 wn -= 1   # a valid up intersect right of P.x

    # wn = 0 only when P is outside the polygon
    if wn == 0:
        return False
    else:
        return True

def isLeft(x1, y1, x2, y2, x3, y3):
    # isLeft(): test if a point is Left|On|Right of an infinite 2D line.
    #   From http://geomalgorithms.com/a01-_area.html
    # Input:  three points P1, P2, and P3
    # Return: >0 for P3 left of the line through P1 to P2
    # =0 for P3 on the line
    # <0 for P3 right of the line
    return ( (x2 - x1) * (y3 - y1) - (x3 - x1) * (y2 - y1) )

def rangeoverlap(a,b):
    """does the range (start1, end1) overlap with (start2, end2)"""
    return max(a) >= min(b) and min(a) <= max(b)
    
def doLineSegmentsIntersect(x1, y1, x2, y2):
    """ Do two line segments intersect."""

    # Check vertical lines

    # Vertical lines, but NOT same X-values
    if x1[0]==x1[1] and x2[0]==x2[1] and x1[0]!=x2[0]:
        return False  # No overlap
    
    # Vertical lines with same X values
    if x1[0]==x1[1] and x2[0]==x2[1] and x1[0]==x2[0]:
        # Check intersection of Y ranges
        I1 = [np.min(y1), np.max(y1)]
        I2 = [np.min(y2), np.max(y2)]
    
        # And we could say that Xa is included into :
        Ia = [max( np.min(y1), np.min(y2) ),
              min( np.max(y1), np.max(y2) )]
    
        # Now, we need to check that this interval Ia exists :
        if rangeoverlap(y1,y2)==False:
            return False  # There is no mutual abcisses        
        else:
            return True   # There is overlap

    # The equation of a line is:
    #
    # f(x) = A*x + b = y
    # For a segment, it is exactly the same, except that x is included on an interval I.
    # 
    # If you have two segments, defined as follow:
    #
    # Segment1 = {(X1, Y1), (X2, Y2)}
    # Segment2 = {(X3, Y3), (X4, Y4)}
    # The abcisse Xa of the potential point of intersection (Xa,Ya) must be contained in both interval I1 and I2, defined as follow:
    I1 = [np.min(x1), np.max(x1)]
    I2 = [np.min(x2), np.max(x2)]
    
    # And we could say that Xa is included into :
    Ia = [max( np.min(x1), np.min(x2) ),
          min( np.max(x1), np.max(x2) )]
    
    # Now, we need to check that this interval Ia exists :
    if rangeoverlap(x1,x2)==False:        
        return False  # There is no mutual abcisses

    # Check that the Y-ranges overlap as well
    if rangeoverlap(y1,y2)==False:        
        return False  # There is no mutual y-value overlap
    
    # So, we have two line formula, and a mutual interval. Your line formulas are:
    # f1(x) = m1*x + b1 = y
    # f2(x) = m2*x + b2 = y
    
    # As we got two points by segment, we are able to determine A1, A2, b1 and b2:
    dx1 = x1[1]-x1[0]
    if dx1==0:
        m1 = np.inf
        b1 = 0
    else:
        m1 = (y1[1]-y1[0])/(x1[1]-x1[0])  # Pay attention to not dividing by zero
        b1 = y1[0]-m1*x1[0]        
    dx2 = x2[1]-x2[0]
    if dx2==0:
        m2 = np.inf
        b2 = 0
    else:
        m2 = (y2[1]-y2[0])/(x2[1]-x2[0])  # Pay attention to not dividing by zero
        b2 = y2[0]-m2*x2[0]
    
    # If the segments are parallel, then m1 == m2:
    if (m1 == m2) and (b1 != b2):
        return False  # Parallel segments
    
    # If the segments are parallel and on top of each other, the m1==m2 and b1==b2
    # we've already required that the x-ranges (abcissas) overlap
    if (m1 == m2) and (b1 == b2):
        return True   # parallel segments on top of each other
    
    # A point (Xa,Ya) standing on both lines must satisfy both formulas f1 and f2:
    # Ya = m1 * Xa + b1
    # Ya = m2 * Xa + b2
    # A1 * Xa + b1 = m2 * Xa + b2

    # Line segment 1 is vertical line 
    if x1[0]==x1[1]:
        Xa = x1[0]
        Ya = m2*Xa+b2
        if rangeoverlap(x1,[Xa]) and rangeoverlap(x2,[Xa]) and \
           rangeoverlap(y1,[Ya]) and rangeoverlap(y2,[Ya]):
            return True
        else:
            return False
    # Line semgent 2 is vertical line
    elif x2[0]==x2[1]:
        Xa = x2[0]
        Ya = m1*Xa+b1
        if rangeoverlap(x1,[Xa]) and rangeoverlap(x2,[Xa]) and \
           rangeoverlap(y1,[Ya]) and rangeoverlap(y2,[Ya]):        
            return True
        else:
            return False
    # Neither are vertical lines
    else:
        Xa = (b2 - b1) / (m1 - m2)
        
    # The last thing to do is check that Xa is included into Ia:
    if ( (Xa < max( np.min(x1), np.min(x2) )) or
         (Xa > min( np.max(x1), np.max(x2) )) ):
        return False  # intersection is out of bound
    else:
        return True

    
# from astroML
def crossmatch(X1, X2, max_distance=np.inf,k=1):
    """Cross-match the values between X1 and X2

    By default, this uses a KD Tree for speed.

    Parameters
    ----------
    X1 : array_like
        first dataset, shape(N1, D)
    X2 : array_like
        second dataset, shape(N2, D)
    max_distance : float (optional)
        maximum radius of search.  If no point is within the given radius,
        then inf will be returned.

    Returns
    -------
    dist, ind: ndarrays
        The distance and index of the closest point in X2 to each point in X1
        Both arrays are length N1.
        Locations with no match are indicated by
        dist[i] = inf, ind[i] = N2
    """
    X1 = np.asarray(X1, dtype=float)
    X2 = np.asarray(X2, dtype=float)

    N1, D = X1.shape
    N2, D2 = X2.shape

    if D != D2:
        raise ValueError('Arrays must have the same second dimension')

    kdt = cKDTree(X2)

    dist, ind = kdt.query(X1, k=k, distance_upper_bound=max_distance)

    return dist, ind

# from astroML, modified by D. Nidever
def xmatch(ra1, dec1, ra2, dec2, dcr=2.0,unique=False,sphere=True):
    """Cross-match angular values between RA1/DEC1 and RA2/DEC2

    Find the closest match in the second list for each element
    in the first list and within the maximum distance.

    By default, this uses a KD Tree for speed.  Because the
    KD Tree only handles cartesian distances, the angles
    are projected onto a 3D sphere.

    This can return duplicate matches if there is an element
    in the second list that is the closest match to two elements
    of the first list.

    Parameters
    ----------
    ra1/dec1 : array_like
        first dataset, arrays of RA and DEC
        both measured in degrees
    ra2/dec2 : array_like
        second dataset, arrays of RA and DEC
        both measured in degrees
    sphere : boolean, optional
        The coordinates are spherical in degrees.  Otherwise, the dcr
          is assumed to be in the same units as the input values.
          Default is True.
    dcr : float (optional)
        maximum radius of search, measured in arcsec.
        This can be an array of the same size as ra1/dec1.

    Returns
    -------
    ind1, ind2, dist: ndarrays
        The indices for RA1/DEC1 (ind1) and for RA2/DEC2 (ind2) of the
        matches, and the distances (in arcsec).
    """
    X1 = np.vstack((ra1,dec1)).T
    X2 = np.vstack((ra2,dec2)).T

    # Spherical coordinates in degrees
    if sphere:
        X1 = X1 * (np.pi / 180.)
        X2 = X2 * (np.pi / 180.)
        if utils.size(dcr)>1:
            max_distance = (np.max(dcr) / 3600) * (np.pi / 180.)
        else:
            max_distance = (dcr / 3600) * (np.pi / 180.)

        # Convert 2D RA/DEC to 3D cartesian coordinates
        Y1 = np.transpose(np.vstack([np.cos(X1[:, 0]) * np.cos(X1[:, 1]),
                                     np.sin(X1[:, 0]) * np.cos(X1[:, 1]),
                                     np.sin(X1[:, 1])]))
        Y2 = np.transpose(np.vstack([np.cos(X2[:, 0]) * np.cos(X2[:, 1]),
                                     np.sin(X2[:, 0]) * np.cos(X2[:, 1]),
                                     np.sin(X2[:, 1])]))

        # law of cosines to compute 3D distance
        max_y = np.sqrt(2 - 2 * np.cos(max_distance))
        k = 1 if unique is False else 10
        dist, ind = crossmatch(Y1, Y2, max_y, k=k)
    
        # convert distances back to angles using the law of tangents
        not_inf = ~np.isinf(dist)
        x = 0.5 * dist[not_inf]
        dist[not_inf] = (180. / np.pi * 2 * np.arctan2(x,
                                np.sqrt(np.maximum(0, 1 - x ** 2))))
        dist[not_inf] *= 3600.0      # in arcsec
    # Regular coordinates
    else:
        k = 1 if unique is False else 10
        dist, ind = crossmatch(X1, X2, np.max(dcr), k=k)
        not_inf = ~np.isinf(dist)
            
    # Allow duplicates
    if unique is False:

        # no matches
        if np.sum(not_inf)==0:
            return [], [], [np.inf]
        
        # If DCR is an array then impose the max limits for each element
        if utils.size(dcr)>1:
            bd,nbd = utils.where(dist > dcr)
            if nbd>0:
                dist[bd] = np.inf
                not_inf = ~np.isinf(dist)
    
        # Change to the output that I want
        ind1 = np.arange(len(ra1))[not_inf]
        ind2 = ind[not_inf]
        mindist = dist[not_inf]

    # Return unique one-to-one matches
    else:

        # no matches
        if np.sum(~np.isinf(dist[:,0]))==0:
            return [], [], [np.inf]
        
        done = 0
        niter = 1
        # Loop until we converge
        while (done==0):

            # If DCR is an array then impose the max limits for each element
            if utils.size(dcr)>1:
                bd,nbd = utils.where(dist[:,0] > dcr)
                if nbd>0:
                    for i in range(nbd):
                        dist[bd[i],:] = np.inf

            # no matches
            if np.sum(~np.isinf(dist[:,0]))==0:
                return [], [], [np.inf]

            # closest matches
            not_inf1 = ~np.isinf(dist[:,0])
            ind1 = np.arange(len(ra1))[not_inf1]
            ind2 = ind[:,0][not_inf1]
            mindist = dist[:,0][not_inf1]
            if len(ind2)==0:
                return [], [], [np.inf]
            index = utils.create_index(ind2)
            # some duplicates to deal with
            bd,nbd = utils.where(index['num']>1)
            if nbd>0:
                torem = []            
                for i in range(nbd):
                    indx = index['index'][index['lo'][bd[i]]:index['hi'][bd[i]]+1]
                    # keep the one with the smallest minimum distance
                    si = np.argsort(mindist[indx])
                    if index['num'][bd[i]]>2:
                        torem += list(indx[si[1:]])    # add list
                    else:
                        torem.append(indx[si[1:]][0])  # add single element
                ntorem = utils.size(torem)
                # For each object that was "removed" and is now unmatched, check the next possible
                # match and move it up in the dist/ind list if it isn't INF
                for i in range(ntorem):
                    # There is a next possible match 
                    if ~np.isinf(dist[torem[i],niter-1]):
                        ind[torem[i],:] = np.hstack( (ind[torem[i],niter:].squeeze(), np.repeat(-1,niter)) )
                        dist[torem[i],:] = np.hstack( (dist[torem[i],niter:].squeeze(), np.repeat(np.inf,niter)) )
                    # All INFs
                    else:
                        ind[torem[i],:] = -1
                        dist[torem[i],:] = np.inf
            else:
                ntorem = 0

            niter += 1
            # Are we done, no duplicates or hit the maximum 10
            if (ntorem==0) | (niter>=10): done=1
                                
    return ind1, ind2, mindist

def dist(x1, y1, x2, y2):
    """ Calculate Euclidian distance between two sets of points."""

    if (utils.size(x1) != utils.size(y1)):
        raise ValueError('x1/y1 must have same number of elements')
    if (utils.size(x2) != utils.size(y2)):
        raise ValueError('x2/y2 must have same number of elements')    

    return np.sqrt( (x1-x2)**2 + (y1-y2)**2 )
    

def sphdist(lon1, lat1, lon2, lat2):
    """Calculate the angular distance between two sets of points.

    Parameters
    ----------
    lon1/lat1 : scalar or array_like
        first dataset, arrays of LON/LAT or RA/DEC
        both measured in degrees
    lon2/lat2 : scalar array_like
        second dataset, arrays of LON/LAT or RA/DEC
        both measured in degrees

    Returns
    -------
    dist: ndarrays
        The angular distance in degrees.
    """

    if (utils.size(lon1) != utils.size(lat1)):
        raise ValueError('lon1/lat1 must have same number of elements')
    if (utils.size(lon2) != utils.size(lat2)):
        raise ValueError('lon2/lat2 must have same number of elements')    
    
    # From this website;
    # http://www2.sjsu.edu/faculty/watkins/sphere.htm

    cosa = np.sin(np.deg2rad(lat1))*np.sin(np.deg2rad(lat2))+np.cos(np.deg2rad(lat1))*np.cos(np.deg2rad(lat2))*np.cos(np.deg2rad(lon1-lon2))
    dist = np.rad2deg(np.arccos(cosa))

    return dist

def lbd2xyz(l,b,d,R0=8.5):
    """
    Convert from LON, LAT and DISTANCE to galactocentric
    cartesian coordinates.

    The L,B and D (distance) are all with respect to the sun
    L and B need to be in degrees and D in kpc.
    The X, Y and Z coordinates are with respect to the Galactic Center
    Z is going up, towards b=90
    X is towards the galactic center from our direction, L=0
    Y is towards the L=90 direction

    Parameters:
    -----------
    l    Galactic longitude in degrees
    b    Galactic latitude in degrees
    d    Distance from sun in kpc
    =R0  Distance of the sun to the galactic center. 8.5 kpc is the default

    Returns:
    -------
    X  The galactocentric cartesian X coordinate in kpc.  Positive X is
          towards the Galactic center or L=0.
    Y  The galactocentric cartesian Y coordinate in kpc.  Positive Y is
          towards L=90.
    Z  The galactocentric cartesian Z coordinate in kpc.  Positive Z is
          towards the North Galactic pole.
    """

    brad = np.deg2rad(np.atleast_1d(b).copy().astype(np.float64))
    lrad = np.deg2rad(np.atleast_1d(l).copy().astype(np.float64))
    dd = np.atleast_1d(d).copy().astype(np.float64)

    x = dd*np.sin(0.5*np.pi-brad)*np.cos(lrad)-R0
    y = dd*np.sin(0.5*np.pi-brad)*np.sin(lrad)
    z = dd*np.cos(0.5*np.pi-brad)

    return x,y,z


def xyz2lbd(x,y,z,R0=8.5):
    """ Convert galactocentric X/Y/Z coordinates to l,b,dist."""

    rho = np.sqrt( (x+R0)**2 + y**2)
    lrad = np.arctan2(y,x+R0)
    brad = 0.5*np.pi - np.arctan2(rho,z)      # this is more straighforward

    brad[brad > 0.5*np.pi] -= np.pi
    brad[brad < -0.5*np.pi] += np.pi

    # This doesn't work if z=0
    #if cos(0.5*!dpi-brad) ne 0.0 then d = zz/cos(0.5*!dpi-brad)
    #if cos(0.5*!dpi-brad) eq 0.0 then d = abs(zz)
    d = np.sqrt( (x+R0)**2 + y**2 + z**2 )
    b = np.rad2deg(brad)
    l = np.rad2deg(lrad)
    l = l % 360

    return l,b,d


class MagellanicStream(coord.BaseCoordinateFrame):
    """
    A Heliocentric spherical coordinate system defined by the Magellanic Stream
    Parameters
    ----------
    representation : `BaseRepresentation` or None
        A representation object or None to have no data (or use the other keywords)
    MSLongitude : `Angle`, optional, must be keyword
        The longitude-like angle corresponding to the Magellanic Stream.
    MSLatitude : `Angle`, optional, must be keyword
        The latitude-like angle corresponding to the Magellanic Stream.
    distance : `Quantity`, optional, must be keyword
        The Distance for this object along the line-of-sight.
    pm_Lambda_cosBeta : :class:`~astropy.units.Quantity`, optional, must be keyword
        The proper motion along the Stream in ``Lambda`` (including the
        ``cos(Beta)`` factor) for this object (``pm_Beta`` must also be given).
    pm_Beta : :class:`~astropy.units.Quantity`, optional, must be keyword
        The proper motion in Declination for this object (``pm_ra_cosdec`` must
        also be given).
    radial_velocity : :class:`~astropy.units.Quantity`, optional, must be keyword
        The radial velocity of this object.

    Developed by Y. Choi

    """
    default_representation = coord.SphericalRepresentation

    frame_specific_representation_info = {
        coord.SphericalRepresentation: [
            coord.RepresentationMapping('lon', 'MSLongitude'),
            coord.RepresentationMapping('lat', 'MSLatitude'),
            coord.RepresentationMapping('distance', 'distance')]#,
    }
    frame_specific_representation_info[coord.UnitSphericalRepresentation] = frame_specific_representation_info[coord.SphericalRepresentation]

MS_PHI = (180 + 8.5 + 90) * u.degree # Euler angles (from Nidever 2010)
MS_THETA = (90 + 7.5) * u.degree
MS_PSI = -32.724214217871349 * u.degree  # anode parameter from gal2mag.pro

D = rotation_matrix(MS_PHI, "z")
C = rotation_matrix(MS_THETA, "x")
B = rotation_matrix(MS_PSI, "z")
A = np.diag([1., 1., 1.])
MS_MATRIX = matrix_product(A, B, C, D)

@frame_transform_graph.transform(coord.StaticMatrixTransform, coord.Galactic, MagellanicStream)
def galactic_to_MS():
    """ Compute the transformation matrix from Galactic spherical to
        Magellanic Stream coordinates.
    """
    return MS_MATRIX

@frame_transform_graph.transform(coord.StaticMatrixTransform, MagellanicStream, coord.Galactic)
def MS_to_galactic():
    """ Compute the transformation matrix from Magellanic Stream coordinates to
        spherical Galactic.
    """
    return matrix_transpose(MS_MATRIX)

#c_icrs = SkyCoord(ra=tmp.ra*u.degree, dec=tmp.dec*u.degree)
#c_ms = c_icrs.transform_to(MagellanicStream)
#ms_l,ms_b = c_ms.MSLongitude.degree, c_ms.MSLatitude.degree #subtract off 360 from ms_l

def gal2mag(glon,glat):
    """ Convert Galactic longitude/latitude to Magellanic Stream longitude/latitude."""
    c_glon = SkyCoord(glon,glat,frame='galactic',unit='deg')
    c_ms = c_glon.transform_to(MagellanicStream)
    mlon,mlat = c_ms.MSLongitude.degree, c_ms.MSLatitude.degree #subtract off 360 from ms_l
    return mlon,mlat

def mag2gal(mlon,mlat):
    """ Convert Magellanic Stream longitude/latitude to Galactic longitude/latitude."""
    c_ms = SkyCoord(mlon,mlat,frame=MagellanicStream,unit='deg')
    c_glon = c_ms.transform_to('galactic')
    glon,glat = c_glon.l.degree, c_glon.b.degree #subtract off 360 from ms_l
    return glon,glat

def wcsfit(wcs,tab,verbose=False):
    """
    Fit the WCS using a catalog of stars with known X/Y and RA/DEC coordinates.

    Parameters
    ----------
    wcs : WCS object
      The WCS object with an estimate of the WCS.
    tab : table
      Catalog with x, y, ra, and dec of matched sources.
    verbose : boolean, optional
      Print informaton to the screen.  Default is False.

    Returns
    -------
    fwcs : WCS object
      WCS with improved paramters.
    
    Example
    -------

    fwcs = wcsfit(wcs,tab)L

    """

    coo = SkyCoord(ra=tab['ra'],dec=tab['dec'],unit='deg')

    def newwcs(pars):
        # pars = [CRVAL1,CRVAL2,CDELT1,CDELT2,PC1_1,PC1_2,PC2_1,PC2_2]
        # pars = [delta_ra (arcsec), delta_dec (arcsec), cdelt1_scale_change (multiplicative)
        #  cdelt2_scale_change (multiplicate), rotation (deg)]
        twcs = copy.deepcopy(wcs)
        twcs.wcs.crval[0] += pars[0]/3600.0
        twcs.wcs.crval[1] += pars[1]/3600.0  
        twcs.wcs.cdelt[0] *= pars[2]
        twcs.wcs.cdelt[1] *= pars[3]
        # Rotation matrix
        # R = [[cos(theta), -sin(theta)], [sin(theta), cos(theta)]]
        rot = np.array([[np.cos(np.deg2rad(pars[4])),-np.sin(np.deg2rad(pars[4]))],
                        [np.sin(np.deg2rad(pars[4])),np.cos(np.deg2rad(pars[4]))] ])
        newpc = np.dot(twcs.wcs.pc,rot)
        twcs.wcs.pc = newpc
        return twcs
    def diffcoords(pars,good=None):
        twcs = newwcs(pars)
        try:
            vcoo = twcs.pixel_to_world(tab['x'],tab['y'])
            diff = coo.separation(vcoo)
            if good is not None:
                meddiff = np.nanmean(diff[good].arcsec)
            else:
                meddiff = np.nanmean(diff.arcsec)
        except:
            meddiff = 999999.
        return meddiff

    def diffcoords_jac(pars):
        delta = [0.2,0.2,0.1,0.1,0.5]
        jac = np.zeros(5,float)
        y0 = diffcoords(pars)
        for i in range(len(pars)):
            tpars = np.array(pars).copy()
            tpars[i] += delta[i]
            y1 = diffcoords(tpars)
            jac[i] = (y1-y0)/delta[i]
        return y0,jac
            
    
    # fit delta_ra (arcsec), delta_dec (arcsec), cdelt1_scale_change (multiplicative)
    #  cdelt2_scale_change (multiplicate), rotation (deg)]


    # Get initial guess of delta_ra and delta_dec using the coordinate ra/dec values
    # and initial WCS coordinates
    vcoo = wcs.pixel_to_world(tab['x'],tab['y'])
    vra = vcoo.ra.deg
    vdec = vcoo.dec.deg
    dra = np.median(tab['ra']-vra)*3600
    ddec = np.median(tab['dec']-vdec)*3600

    # Estimate the rotation
    vx,vy = wcs.world_to_pixel(coo)
    coefx,absdevx = ladfit.ladfit(tab['y'],tab['x']-vx)
    coefy,absdevy = ladfit.ladfit(tab['x'],tab['y']-vy)    
    rot = np.mean([coefx[1],-coefy[1]])

    # Do the fit
    estimates = [dra,ddec,1.0,1.0,rot]    
    bounds = len(estimates)*[[-np.inf,np.inf]]
    res1 = minimize(diffcoords,estimates,bounds=bounds)
    pars1 = res1.x
    
    # Remove outliers and refit
    twcs = newwcs(pars1)
    vcoo = twcs.pixel_to_world(tab['x'],tab['y'])
    diff = coo.separation(vcoo)
    meddiff = np.nanmedian(diff.arcsec)
    sigdiff = utils.mad(diff.arcsec)
    good, = np.where(diff.arcsec < (meddiff+3.5*sigdiff))
    noutlier = len(tab)-len(good)
    if noutlier>0:
        if verbose:
            print('Rejecting '+str(noutlier)+' outlier(s)')
        res = minimize(diffcoords,pars1,args=good,bounds=bounds)
        pars = res.x
    else:
        res = res1
        pars = pars1
                  
    fwcs = newwcs(pars)
    
    if verbose:
        print('--- Original WCS ---')
        print(wcs)
        print(' ')
        print('--- Final WCS ---')
        print(fwcs)
        print(' ')
        resid0 = diffcoords([0.0,0.0,1.0,1.0,0.0])
        resid = diffcoords(pars,good)
        print('Original mean residuals: {:.3f} arcsec'.format(resid0))
        print('Final mean residuals   : {:.3f} arcsec'.format(resid))
        
    return fwcs



#!/usr/bin/env python
#
# JOB_DAEMON.PY - Simple batch job manager.
#

from __future__ import print_function

__authors__ = 'David Nidever <dnidever@montana.edu>'
__version__ = '20181010'  # yyyymmdd

import os
import sys
import numpy as np
import warnings
import socket
import time
#import shutil
import subprocess
#import logging
import tempfile
from . import utils as dln

# Ignore these warnings, it's a bug
warnings.filterwarnings("ignore", message="numpy.dtype size changed")
warnings.filterwarnings("ignore", message="numpy.ufunc size changed")
        

def mkstatstr(n=None):
    """ This returns the stat structure schema or an instance of the stat structure."""
    dtype = np.dtype([('jobid',np.str,20),('name',np.str,100),('user',np.str,100),('timeuse',np.str,100),('status',np.str,10),('queue',np.str,20)])
    if n is None:
        return dtype
    else:
        statstr = np.zeros(n,dtype=dtype)    
        return statstr


def mkjobstr(n=None):
    """ This returns the job structure schema or an instance of the job structure."""
    dtype = np.dtype([('host',np.str,20),('jobid',(np.str,100)),('input',np.str,1000),('dir',np.str,500),('name',np.str,100),('scriptname',np.str,200),
                      ('logfile',np.str),('submitted',np.bool),('done',np.bool),('begtime',np.float64),('endtime',np.float64),('duration',float)])
    if n is None:
        return dtype
    else:
        jobstr = np.zeros(n,dtype=dtype)
        jobstr['submitted'] = False
        jobstr['done'] = False
        return jobstr


def mkrunbatch():
    curdir = os.getcwd()
    batchfile = os.path.join(curdir,'runbatch')
    if os.path.exists(batchfile) is False:
        lines = []
        lines.append("if test $# -eq 0\n")
        lines.append("then\n")
        lines.append("  echo 'Syntax - runbatch program'\n")
        lines.append("else\n")
        lines.append("  echo 'Log file: '$1'.log'\n")
        lines.append("  ( nohup  $1 > $1.log 2>&1 ) &\n")
        lines.append("  echo 'JOBID='$!\n")
        lines.append("fi\n")
        dln.writelines(batchfile,lines,overwrite=True,raw=True)
        os.chmod(batchfile,0o755)
    return batchfile


def mkidlbatch():
    curdir = os.getcwd()
    batchfile = os.path.join(curdir,'idlbatch')
    if os.path.exists(batchfile) is False:
        try:
            out = subprocess.check_output(['which','idl'],stderr=subprocess.STDOUT,shell=False)
        except subprocess.CalledProcessError:
            raise Exception("IDL program not available")
        idlprog = out.decode().strip()
        if os.path.exists(idlprog) is False:
            raise Exception("IDL program "+idlprog+" not found")
        lines = []
        lines.append("if test $# -eq 0\n")
        lines.append("then\n")
        lines.append("  echo 'Syntax - idlbatch idl.batch'\n")
        lines.append("else\n")
        lines.append("  echo 'Log file: '$1'.log'\n")
        lines.append("  ( nohup "+idlprog+" < $1 > $1.log 2>&1 ) &\n")
        lines.append("  echo 'JOBID='$!\n")
        lines.append("fi\n")
        dln.writelines(batchfile,lines,overwrite=True,raw=True)
        os.chmod(batchfile,0o755)
    return batchfile


def check_diskspace(indir=None,updatestatus=False):
    """ This checks if there's enough free disk space. """
    if indir is None: raise ValueError("Must give directory")
    statvfs = os.statvfs(indir)
    available = statvfs.f_frsize * statvfs.f_bavail / 1e6  # in MB
    if updatestatus: print('Disk Space: '+str(available[0]),' MB available')
    # Not enough disk space available
    if available<100.:
        raise Exception('NOT enough disk space available')


def check_killfile(jobs=None,hyperthread=True):
    """ This checks for a kill file and if found kills all the active jobs. """
    if jobs is None: raise ValueError("No jobs structure input")
    killfile = 'killjobs'
    if os.path.exists(killfile) is True:
        sub, = np.where((jobs['submitted']==1) & (jobs['done']==0))
        nsub = len(sub)
        print('Kill file found.  Killing all '+str(nsub)+' job(s)')
        for i in range(nsub):
            # Killing the job
            print('Killing '+jobs['name'][sub[i]]+'  JobID='+jobs['jobid'][sub[i]])
            if hyperthread is False:
                out = subprocess.check_output(['qdel',jobs['jobid'][sub[i]]],stdout=sf,stderr=subprocess.STDOUT,shell=False)
            else:
                # Using negative PID will also kill subprocesses
                try:
                    out = asubprocess.check_output(['kill','-9','-'+jobs['jobid'][sub[i]]],stderr=subprocess.STDOUT,shell=False)
                except:
                    out = asubprocess.check_output(['kill','-9',jobs['jobid'][sub[i]]],stderr=subprocess.STDOUT,shell=False)                    
        # Remove the kill file
        print('Deleting kill file "'+killfile+'"')
        os.remove(killfile)
        return True
    return False


def makescript(inp=None,indir=None,name=None,prefix=None,hyperthread=True,idle=False):
    """This makes job scripts for the job_daemon program.

    Parameters
    ----------
    inp : string list or array
          The command to execute.  Can be an array.  Must input absolute path names.
          idlbatch will be used if it's an IDL command.  If the command is a
          series of unix commands separated by commas then these
          will be put on separate lines.
    indir : string list or array
          The directory to put the job script in.
    name : string list or array, optional
          The name to call the job script (without the '.sh' or '.batch' ending).
          If this is not provided then an autogenerated name will
          be made and returned.
    prefix : string, optional
          The prefix to use for the job script name. "pr" by default.
    hyperthread : bool, optional
          Not on a job server but one with multiple hyperthreaded
                   processors.
    idle : bool, optional
          This is an IDL command, otherwise a SHELL command.

    Returns
    -------
    scriptname : string array
               The absolute names of the scripts.
    Job scripts output to the directories and with the names specified.


    Example
    -------

    .. code-block:: python

        scriptname = makescript(inp,dir,name,hyperthread=True)

    """
  
    # Not enough inputs
    if (inp is None):
        raise ValueError('No input given')

    ninp = dln.size(inp)
    ndir = dln.size(indir)
    nname = dln.size(name)

    # Not enough directories input
    if (ndir>0) & (ndir!=ninp):
        raise ValueError('INPUT and DIRECTORIES are of different size')

    # Current directory
    curdir = os.getcwd()

    # No directories input
    if ndir==0: indir = np.repeat(curdir,ninp)
    if ndir==1: indir = np.repeat(indir,ninp)    # multiple commands in same dir

    # Construct names
    if (nname==0):
        name = np.zeros(ninp,dtype=(np.str,200))
        if prefix is not None:
            pre = dln.first_el(prefix)
        else:
            pre = 'job'
        for i in range(ninp):
            tid,tfile = tempfile.mkstemp(prefix=pre,dir=indir[i])
            os.close(tid)   # mkstemp opens the file, close it
            name[i] = os.path.basename(tfile)

    # Make scriptnames
    scriptname = np.array(dln.strjoin(dln.pathjoin(indir,name),'.sh'),ndmin=1)
    # Script loop
    for i,input1 in enumerate(np.array(inp,ndmin=1)):
        base = str(name[i])
        sname = str(indir[i])+'/'+base+'.sh'

        #------
        # PBS
        #------
        if hyperthread is False:
            # IDL command
            if idle is True:
                # Making an IDL batch file
                bname = str(indir[i])+'/'+base+'.batch'
                dln.writelines(bname,input1,overwrite=True,raw=True)
                # The execution command
                cmd = 'idl < '+base+'.batch'
            # SHELL command
            else:
                # The execution command
                cmd = input1
                # If there are commas in the line then break it up into multiple lines
                if cmd.find(';') != -1:
                    cmd = cmd.replace(';','\n;')
                    cmd = cmd.split('j')

            # Make the command
            #----------------------
            lines = []
            lines.append('#!/bin/sh\n')
            lines.append('#PBS -l nodes=1:ppn=1\n')
            lines.append('#PBS -l walltime=96:00:00\n')
            lines.append('#PBS -o '+base+'.report.out\n')
            lines.append('#PBS -e '+base+'.error.out\n')
            #lines.append('#PBS -M dln5q@virginia.edu\n')
            lines.append('#PBS -V\n')
            lines.append('\n')
            lines.append('echo Running on host `hostname`\n')
            lines.append('echo Time is `date`\n')
            lines.append('echo "Nodes used for this job:"\n')
            lines.append('echo "------------------------"\n')
            lines.append('cat $PBS_NODEFILE\n')
            lines.append('echo "------------------------"\n')
            lines.append('\n')
            lines.append('cd '+indir[i]+'\n')
            for j in range(len(cmd)): lines.append(cmd[j]+'\n')
            lines.append('\n')
            lines.append('# print end time\n')
            lines.append('echo\n')
            lines.append('echo "Job Ended at `date`"\n')
            lines.append('echo\n')
            # Writing the file
            dln.writelines(scriptname[i],lines,overwrite=True,raw=True)
            # Print info
            print('PBS script written to: '+str(scriptname[i]))

        #----------------
        # Hyperthreaded
        #----------------
        else:
            # Just make batch file
            # treat shell and idl the same
            # The execution command
            cmd = input1
            # If there are commas in the line then break it up into multiple lines
            if cmd.find(';') != -1:
                cmd = cmd.replace(';','\n;')
                cmd = cmd.split(';')
            # IDL files should end in .batch
            if idle is True: scriptname[i] = str(indir[i])+'/'+base+'.batch'
            # Writing the file
            dln.writelines(scriptname[i],cmd,overwrite=True,raw=True)
            # Make SHELL scripts executable
            if idle is False: os.chmod(scriptname[i],0o755)
            # Print info
            print('HYPERTHREAD script written to: '+str(scriptname[i]))

    # Erase the temporary files that mkstemp makes
    dln.remove(dln.pathjoin(indir,name),allow=True)

    if dln.size(scriptname)==1: scriptname=scriptname[0]
    return scriptname


def submitjob(scriptname=None,indir=None,hyperthread=True,idle=False):
    """ This submits new jobs.

    Parameters
    ----------
    scriptname : string
           Name of the script to run.
    indir : string, optional
           Directory to change to before script is executed.
    hyperthread : bool, optional
          Not on a job server but one with multiple hyperthreaded
             processors.  By default, this is True.
    idle : bool, optional
         This is an IDL program.  By default his is False.

    Returns
    -------
    jobid : string
         The job ID for this job.
    logfile : string
         Name of the output log file for this job/script.
    The script `scriptname` will be run.

    Example
    -------

    .. code-block:: python

        jobid,logfile = submitjob(scriptname,indir,hyperthread=True)

    """
    if scriptname is None: raise ValueError("scriptname must be input")

    curdir = os.getcwd()
    # Submitting the job
    if hyperthread is False:
        try:
            out = subprocess.check_output('qsub '+scriptname,stderr=subprocess.STDOUT,shell=True)
        except subprocess.CalledProcessError:
            raise Exception("Problem submitting PBS job")
        jobid = dln.first_el(out)
        logfile = scriptname+'.log'
    else:
        if idle is True:
            batchprog = mkidlbatch()
        else:
            batchprog = mkrunbatch()
        if indir is not None: os.chdir(indir)
        try:
            out = subprocess.check_output(batchprog+' '+scriptname,stderr=subprocess.STDOUT,shell=True)
        except:
            raise Exception("Problem submitting shell job")
        if indir is not None: os.chdir(curdir)
        # Get the JOBID
        out = out.decode().split('\n')
        jobid_ind = dln.grep(out,'^JOBID=',index=True)
        njobid_ind = len(jobid_ind)
        jobid = out[jobid_ind[0]].split('=')[1]
        jobid = jobid.strip()
        # Get the logfile
        logfile_ind = dln.grep(out,'^Log file: ',index=True)
        nlogfile_ind = len(logfile_ind)
        logfile = out[logfile_ind[0]].split(':')[1]
        logfile = logfile.strip()
    # Printing info
    print('Submitted '+scriptname+'  JobID='+jobid)

    return jobid, logfile


def getstat(jobid=None,hyperthread=True):
    """Get the status of jobs

    This checks the status of jobs for the job_daemon program.
    If no jobs are found in the queue then an empty
    statstr structure is returned.

    Parameters
    ----------
    jobid : string or int
          Specific JOBID to check.
    hyperthread : bool, optional
          Not on a PBS machine but one with multipe hyperthreaded
          processors running simultaneously.  Default is True.

    Results
    -------
    statstr : numpy structured array
          Status structure.

    Example
    -------

    .. code-block:: python

        stat = getstat(jobid)

    """
    if jobid is None: raise ValueError("Must input jobid")
    njobid = dln.size(jobid)

    # PBS
    #--------
    if hyperthread is False:
        if njobid>0: 
            out = subprocess.check_output(['qstat',jobid[0]],stderr=subprocess.STDOUT,shell=False)
        else:
            out = subprocess.check_output(['qstat'],stderr=subprocess.STDOUT,shell=False)
        nout = np.sum(out.strip() != '')
        gd = dln.grep(out,'^'+jobid,index=True)
        ngd = len(gd)
        if ngd>0:
            statlines = out[gd[0]]
            nstat = len(statlines)
        else:
            statlines = None
        # Some jobs in queue
        if statlines is not None:
            arr = statlines
            arr = arr.split(' ')
            statstr = mkstatstr(nstat)
            statstr['jobid'] = arr[0]
            statstr['name'] = arr[1]
            statstr['user'] = arr[2]
            statstr['timeuse'] = arr[3]
            statstr['status'] = arr[4]
            statstr['queue'] = arr[5]
        # No jobs in queue
        else:
            statstr = mkstatstr(1)

    # Hyperthreaded.  Need a jobid
    #-----------------------------
    else:
        # No JOBID input
        if njobid==0:
            print('Need JOBID with /hyperthread')
            return mkstatstr(1)

        try:
            out = subprocess.check_output(['ps','-o','pid,user,etime,command','-p',str(dln.first_el(jobid))],
                                          stderr=subprocess.STDOUT,shell=False)
        except:
            statstr = mkstatstr(1)
            statstr['jobid'] = jobid
            statstr['queue'] = 'hyperthread'
            return statstr
        # can put in the column that you want
        # ps -o etime -p jobid
        out = dln.strsplit(out.decode(),'\n')
        out = dln.strip(out)
        gd = dln.grep(out,'^'+str(jobid),index=True)
        ngd = len(gd)
        if ngd>0:
            statlines = np.array(out,ndmin=1)[gd[0]]
        else:
            statlines = None
        # Some jobs in queue
        if statlines is not None:
            arr = statlines.split()
            statstr = mkstatstr(1)
            statstr['jobid'] = arr[0]
            statstr['user'] = arr[1]
            # CAN'T get the name.
            statstr['timeuse'] = arr[2]
            statstr['status'] = 'R'
            statstr['queue'] = 'hyperthread'
        # No jobs in queue
        else:
            statstr = mkstatstr(1)
            statstr['jobid'] = jobid
            statstr['queue'] = 'hyperthread'
    return statstr


def checkjobs(jobs=None,hyperthread=True):
    """ Check on the status of the running jobs"""
    if jobs is None: raise ValueError("jobs must be input")
    sub, = np.where((jobs['submitted']==True) & (jobs['done']==False))
    nsub = len(sub)
    nfinished = 0
    for i in range(nsub):
        # Checking status
        jobid = jobs[sub[i]]['jobid']
        statstr = getstat(jobid,hyperthread=hyperthread)
        # Job done
        if statstr['status']=='':
            if nfinished==0: print('')
            print(time.ctime()+'  Input '+str(sub[i]+1)+' '+jobs['name'][sub[i]]+' JobID='+jobs['jobid'][sub[i]]+' FINISHED')
            jobs['done'][sub[i]] = True
            jobs['endtime'][sub[i]] = time.time()/3600/24   # in days
            jobs['duration'][sub[i]] = (jobs['endtime'][sub[i]] - jobs['begtime'][sub[i]])*3600*24   # in sec
            nfinished += 1
            # Check for errors as well!! and put in jobs structure
    return jobs


def status_update(jobs=None):
    """ Print out the status update."""
    if jobs is None: raise ValueError("jobs must be input")
    njobs = dln.size(jobs)                                                   # Number of total jobs
    n_inqueue = np.sum((jobs['submitted']==True) & (jobs['done']==False))    # Number of jobs still in queue  
    n_nosubmit = np.sum(jobs['submitted']==False)                            # Number of jobs left to do 
    n_finished = np.sum(jobs['done']==True)                                  # Number of jobs finished 
    # Print the status
    print('')
    print(time.ctime())
    print(('Jobs Summary: %d total, %d finished, %d running, %d left') % (njobs,n_finished,n_inqueue,n_nosubmit))


def job_daemon(inp=None,dirs=None,inpname=None,nmulti=4,prefix="job",hyperthread=True,idle=False,
               waittime=0.2,statustime=60):
    """This program is a simple batch job manager

    NOTE:  If you want to "kill" all of the jobs create a file "killjobs"
    in the same directory that JOB_DAEMON is being run in and all of the
    PBS jobs will be killed.

    Parameters
    ----------
    inp : string array/list
          A string array with the commands (i.e. jobs) to be run.
    dirs : string array/list
          The directories in which the commands are to be run.
    inpname : string array/list, optional
           The name to be used for the script (without the path or
           the .sh/.batch ending.  This is normally autogenerated.
    nmulti : int, optional
           How many nodes to run these jobs on.  Default is 4.
    prefix : string
          The prefix for the script names.  Default is "job".
    hyperthread : bool, optional
           Not on a PBS server but one that has multiple processors
           hyperthreaded.  Run multiple jobs at the same time on
           the same server.  Default is True.
    idle : bool, optional
         This is an IDL command, otherwise a SHELL command.  Default is False.
    statustime : float or int, optional
           The time between status updates.  However, the status
           will always be updated if something has actually changed.
           Default is 60.
    waittime : float or int, optional
           Time to wait between checking the running jobs.  Default is 0.2 sec.

    Results
    -------
    jobs : numpy structured array
          The jobs structure with information on the JOBID, script name, etc.
    Jobs are run in a batch mode.

    Example
    -------

    .. code-block:: python

        jobs = job_daemon(input,dirs,hyperthread=True,nmulti=5)

    """

    # How many input lines
    if inp is None:
        raise ValueError("Nothing input")
    ninp = dln.size(inp)

    # Current directory
    curdir = os.getcwd()

    # Checking DIRS array
    if dirs is None:
        dirs = np.repeat(curdir,ninp)
    else:
        ndirs = dln.size(dirs)
        if ndirs!=ninp:
            raise ValueError('DIRS array must be same size as INPUT')            
        if ndirs==1: dirs = np.repeat(dirs,ninp)

    # Check INPNAME array
    if inpname is not None:
        ninpname = len(inpname)
        if ninpname!=ninp:
            raise ValueError('INPNAME array must be same size as INPUT')

    # Defaults
    if waittime<0.1: waittime=0.1
    if statustime<1: statustime=1

    # Host name
    hostname = socket.gethostname()
    host = hostname.split('.')[0]

    # Which IDL are we using?
    if idle is True:
        try:
            out = subprocess.check_output(['which','idl'],stderr=subprocess.STDOUT,shell=False)
        except subprocess.CalledProcessError:
            raise Exception("IDL program not available")
        idlprog = out.strip()
        if os.path.exists(idlprog) is False:
            raise Exception("IDL program "+idlprog+" not found")

    print('---------------------------------')
    print(' RUNNING JOB_DAEMON for '+str(ninp)+' JOB(S)')
    print('---------------------------------')
    print('Host='+host)
    print('Nmulti='+str(nmulti))
    
    t0 = time.time()
    timesincelaststatus = time.time()  # initializing the update time

    #--------
    # DAEMON
    #--------
    # -Keep submitting jobs until nmulti is reached
    # -Check every minute or so to see how many jobs are still
    #  running.  If it falls below nmulti and more jobs are left then
    #  submit more jobs
    # -Don't return until all jobs are done.

    # Initialize the "jobs" structure
    # id will be the ID from Pleione
    jobs = mkjobstr(ninp)
    jobs['host'] = host
    jobs['input'] = inp
    njobs = ninp

    # Loop until all jobs are done
    # On each loop check the pleione queue and figure out what to do
    count = np.longlong(0)
    endflag = 0
    while (endflag==0):
        # Status update
        dtstatus_sec = time.time()-timesincelaststatus
        if (dtstatus_sec>statustime):
            updatestatus = True
            timesincelaststatus = time.time()
        else:
            updatestatus = False
  
        # Check disk space
        check_diskspace(dln.first_el(dirs))
        # Check for kill file
        if check_killfile(jobs) is True: return jobs

        # Check status of running jobs
        #-----------------------------
        jobs = checkjobs(jobs,hyperthread=hyperthread)

        # Status update
        #---------------
        if updatestatus is True: status_update(jobs)

        # Submit new jobs
        #----------------
        n_inqueue = np.sum((jobs['submitted']==True) & (jobs['done']==False))  # Number of jobs still in queue  
        n_nosubmit = np.sum(jobs['submitted']==False)                            # Number of jobs left to do 
        nnew = dln.limit(nmulti-n_inqueue,0,n_nosubmit)
        if (nnew>0):
            # Get the indices of new jobs to be submitted
            nosubmit, = np.where(jobs['submitted']==False)
            newind = nosubmit[0:nnew]
            # Update immediately if there are new jobs to submit
            print('')
            print(time.ctime())
            print('Updating Queue: '+str(n_inqueue)+' JOB(S) running, out of '+str(nmulti)+' Maximum. Submitting '+str(nnew)+' more job(s)')

            # Loop through the new submits
            for i in range(nnew):
                print('')
                cmd = jobs[newind[i]]['input']
                if idle is True: cmd = 'IDL>'+str(cmd)
                print('Input '+str(newind[i]+1)+'  Command: >>'+str(cmd)+'<<')
                # Make script
                name = None
                if inpname is not None: name = inpname[newind[i]]  # use input name      
                scriptname = makescript(jobs[newind[i]]['input'],indir=dirs[newind[i]],name=name,
                                        prefix=prefix,hyperthread=hyperthread,idle=idle)
                name = os.path.basename(os.path.splitext(scriptname)[0])
                # Submitting the job
                jobid, logfile = submitjob(scriptname,dirs[newind[i]],hyperthread=hyperthread,idle=idle)
                # Updating the jobs structure
                jobs['submitted'][newind[i]] = True
                jobs['jobid'][newind[i]] = jobid
                jobs['name'][newind[i]] = name
                jobs['dir'][newind[i]] = dirs[newind[i]]
                jobs['scriptname'][newind[i]] = scriptname
                jobs['logfile'][newind[i]] = logfile
                jobs['begtime'][newind[i]] = time.time()/3600/24   # in days

        # Are we done?
        #-------------
        ndone = np.sum(jobs['done']==True)
        if (ndone==njobs): endflag=1
        # Wait a bit
        #--------------
        if (endflag==0): time.sleep(waittime)
        # Increment the counter
        count += 1

    print('DONE')
    print('dt = '+str(time.time()-t0)+' sec')
    return jobs



# gap.py
# (c) 2013 Mikael Vejdemo-Johansson
# BSD License
#
# SciPy function to compute the gap statistic for evaluating k-means clustering.
# Gap statistic defined in
# Tibshirani, Walther, Hastie:
#  Estimating the number of clusters in a data set via the gap statistic
#  J. R. Statist. Soc. B (2001) 63, Part 2, pp 411-423

import scipy
import scipy.cluster.vq
import scipy.spatial.distance
dst = scipy.spatial.distance.euclidean

def gap(data, refs=None, nrefs=20, ks=range(1,11)):
    """
    Compute the Gap statistic for an nxm dataset in data.
    Either give a precomputed set of reference distributions in refs as an (n,m,k) scipy array,
    or state the number k of reference distributions in nrefs for automatic generation with a
    uniformed distribution within the bounding box of data.
    Give the list of k-values for which you want to compute the statistic in ks.
    """
    shape = data.shape
    if refs==None:
        tops = data.max(axis=0)
        bots = data.min(axis=0)
        dists = scipy.matrix(scipy.diag(tops-bots))
	
        rands = scipy.random.random_sample(size=(shape[0],shape[1],nrefs))
        for i in range(nrefs):
            rands[:,:,i] = rands[:,:,i]*dists+bots
    else:
        rands = refs

    gaps = scipy.zeros((len(ks),))
    for (i,k) in enumerate(ks):
        (kmc,kml) = scipy.cluster.vq.kmeans2(data, k)
        disp = sum([dst(data[m,:],kmc[kml[m],:]) for m in range(shape[0])])

        refdisps = scipy.zeros((rands.shape[2],))
        for j in range(rands.shape[2]):
            (kmc,kml) = scipy.cluster.vq.kmeans2(rands[:,:,j], k)
            refdisps[j] = sum([dst(rands[m,:,j],kmc[kml[m],:]) for m in range(shape[0])])
        #gaps[i] = scipy.log(scipy.mean(refdisps))-scipy.log(disp)
        gaps[i] = scipy.mean(scipy.log(refdisps))-scipy.log(disp)
    return gaps




#!/usr/env python

# Imports
import numpy as np
from scipy.special import erf
from scipy.optimize import curve_fit, least_squares
from astropy.io import fits
from scipy import ndimage
from scipy.interpolate import interp1d
from scipy.special import wofz
from . import utils


def trace(im,yestimate=None,yorder=2,sigorder=4,step=10):
    """ Trace the spectrum.  Spectral dimension is assumed to be on the horizontal axis."""
    ny,nx = im.shape
    if yestimate is None:
        ytot = np.sum(im,axis=1)
        yestimate = np.argmax(ytot)
    # Smooth in spectral dimension
    # a uniform (boxcar) filter with a width of 50
    smim = ndimage.uniform_filter1d(im, 50, 1)
    nstep = nx//step
    # Loop over the columns in steps and fit Gaussians
    tcat = np.zeros(nstep,dtype=np.dtype([('x',float),('pars',float,4)]))
    for i in range(nstep):
        pars,cov = dln.gaussfit(y[yestimate-10:yestimate+10],im[yestimate-10:yestimate+10,step*i+step//2])
        tcat['x'][i] = step*i+step//2
        tcat['pars'][i] = pars
    # Fit polynomail to y vs. x and gaussian sigma vs. x
    ypars = np.polyfit(tcat['x'],tcat['pars'][:,1],yorder)
    sigpars = np.polyfit(tcat['x'],tcat['pars'][:,2],sigorder)
    # Model
    mcat = np.zeros(nx,dtype=np.dtype([('x',float),('y',float),('sigma',float)]))
    xx = np.arange(nx)
    mcat['x'] = xx
    mcat['y'] = np.poly1d(ypars)(xx)
    mcat['sigma'] = np.poly1d(sigpars)(xx)
    return tcat, ypars, sigpars,mcat

def boxcar(im):
    """ Boxcar extract the spectrum"""
    ny,nx = im.shape
    ytot = np.sum(im,axis=1)
    yest = np.argmax(ytot)
    # Background subtract
    yblo = int(np.maximum(yest-50,0))
    ybhi = int(np.minimum(yest+50,ny))
    med = np.median(im[yblo:ybhi,:],axis=0)
    medim = np.repeat(med,ny).reshape(ny,nx)
    subim = im.copy()-medim
    # Sum up the flux
    ylo = int(np.maximum(yest-20,0))
    yhi = int(np.minimum(yest+20,ny))
    flux = np.sum(subim[ylo:yhi,:],axis=0)
    return flux

def linefit(x,y,initpar,bounds,err=None):
    # Fit Gaussian profile to data with center and sigma fixed.
    # initpar = [height, center, sigma, constant offset]
    cen = initpar[1]
    sigma = initpar[2]
    def gline(x, amp, const=0):
        """1-D gaussian: gaussian(x, amp, cen, sig)"""
        return amp * np.exp(-(x-cen)**2 / (2*sigma**2)) + const
    line_initpar = [initpar[0],initpar[3]]
    lbounds, ubounds = bounds
    line_bounds = ([lbounds[0],lbounds[3]],[ubounds[0],ubounds[3]])
    return curve_fit(gline, x, y, p0=line_initpar, bounds=line_bounds, sigma=err)


def extract(im,imerr=None,mcat=None,nobackground=False):
    """ Extract a spectrum"""
    ny,nx = im.shape
    x = np.arange(nx)
    y = np.arange(ny)
    # No trace information input, get it
    if mcat is None:
        tcat,ypars,sigpars,mcat=trace(im)
    # Loop over the columns and get the flux using the trace information
    cat = np.zeros(nx,dtype=np.dtype([('x',int),('pars',float,2),('perr',float,2),
                                      ('flux',float),('fluxerr',float)]))
    for i in range(nx):
        line = im[:,i].flatten()
        if imerr is not None:
            lineerr = imerr[:,i].flatten()
        else:
            lineerr = np.ones(len(line))   # unweighted
        # Fit the constant offset and the height of the Gaussian
        #  fix the central position and sigma
        ycen = mcat['y'][i]
        ysigma = mcat['sigma'][i]
        ht0 = np.maximum(line[int(np.round(ycen))],0.01)
        initpar = [ht0,ycen,ysigma,np.median(line)]
        if nobackground is True:
            initpar = [ht0,ycen,ysigma,0]
        # Only fit the region right around the peak
        y0 = int(np.maximum(ycen-50,0))
        y1 = int(np.minimum(ycen+50,ny))
        bnds = ([0,ycen-1e-4,ysigma-1e-4,0],[1.5*ht0,ycen,ysigma,1.5*initpar[3]])
        if nobackground is True:
            bnds = ([0,ycen-1e-4,ysigma-1e-4,0],[1.5*ht0,ycen,ysigma,0.1])
        pars,cov = linefit(y[y0:y1],line[y0:y1],initpar=initpar,bounds=bnds,err=lineerr[y0:y1])
        perr = np.sqrt(np.diag(cov))
        # Gaussian area = ht*wid*sqrt(2*pi)
        flux = pars[0]*ysigma*np.sqrt(2*np.pi)
        fluxerr = perr[0]*ysigma*np.sqrt(2*np.pi)
        cat['x'][i] = i
        cat['pars'][i] = pars
        cat['perr'][i] = perr
        cat['flux'][i] = flux
        cat['fluxerr'][i] = fluxerr
    return cat

def emissionlines(spec,thresh=None):
    """Measure the emission lines in an arc lamp spectrum. """
    nx = len(spec)
    x = np.arange(nx)
    
    # Threshold
    if thresh is None:
        thresh = np.min(spec) + (np.max(spec)-np.min(spec))*0.05
    
    # Detect the peaks
    sleft = np.hstack((0,spec[0:-1]))
    sright = np.hstack((spec[1:],0))
    peaks, = np.where((spec>sleft) & (spec>sright) & (spec>thresh))
    npeaks = len(peaks)
    print(str(npeaks)+' peaks found')
    
    # Loop over the peaks and fit them with Gaussians
    gcat = np.zeros(npeaks,dtype=np.dtype([('x0',int),('x',float),('xerr',float),('pars',float,4),('perr',float,4),
                                           ('flux',float),('fluxerr',float)]))
    resid = spec.copy()
    gmodel = np.zeros(nx)
    for i in range(npeaks):
        x0 = peaks[i]
        xlo = np.maximum(x0-6,0)
        xhi = np.minimum(x0+6,nx)
        initpar = [spec[x0],x0,1,0]
        bnds = ([0,x0-3,0.1,0],[1.5*initpar[0],x0+3,10,1e4])
        pars,cov = dln.gaussfit(x[xlo:xhi],spec[xlo:xhi],initpar,bounds=bnds,binned=True)
        perr = np.sqrt(np.diag(cov))
        gmodel1 = dln.gaussian(x[xlo:xhi],*pars)
        gmodel[xlo:xhi] += (gmodel1-pars[3])
        resid[xlo:xhi] -= (gmodel1-pars[3])
        # Gaussian area = ht*wid*sqrt(2*pi)
        flux = pars[0]*pars[2]*np.sqrt(2*np.pi)
        fluxerr = perr[0]*pars[2]*np.sqrt(2*np.pi)
        gcat['x0'][i] = x0
        gcat['x'][i] = pars[1]
        gcat['xerr'][i] = perr[1]
        gcat['pars'][i] = pars
        gcat['perr'][i] = perr
        gcat['flux'][i] = flux
        gcat['fluxerr'][i] = fluxerr
        
    return gcat, gmodel


def continuum(spec,bin=50,perc=60,norder=4):
    """ Derive the continuum of a spectrum."""
    nx = len(spec)
    x = np.arange(nx)
    # Loop over bins and find the maximum
    nbins = nx//bin
    xbin1 = np.zeros(nbins,float)
    ybin1 = np.zeros(nbins,float)
    for i in range(nbins):
        xbin1[i] = np.mean(x[i*bin:i*bin+bin])
        ybin1[i] = np.percentile(spec[i*bin:i*bin+bin],perc)
    # Fit polynomial to the binned values
    coef1 = np.polyfit(xbin1,ybin1,norder)
    cont1 = np.poly1d(coef1)(x)
    
    # Now remove large negative outliers and refit
    gdmask = np.zeros(nx,bool)
    gdmask[(spec/cont1)>0.8] = True
    xbin = np.zeros(nbins,float)
    ybin = np.zeros(nbins,float)
    for i in range(nbins):
        xbin[i] = np.mean(x[i*bin:i*bin+bin][gdmask[i*bin:i*bin+bin]])
        ybin[i] = np.percentile(spec[i*bin:i*bin+bin][gdmask[i*bin:i*bin+bin]],perc)
    # Fit polynomial to the binned values
    coef = np.polyfit(xbin,ybin,norder)
    cont = np.poly1d(coef)(x)
    
    return cont,coef
